# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: htpass.py
# Project: auth
# Author: Brian Cherinka
# Created: Wednesday, 21st October 2020 2:52:25 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Wednesday, 21st October 2020 2:52:25 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import pathlib
import warnings
from sdss_brain import cfg_params
from sdss_brain.exceptions import BrainError

try:
    import passlib
except ImportError:
    passlib = None
else:
    from passlib.apache import HtpasswdFile


class Htpass(object):
    """ Class representing a htpasswd file

    This class represents a local .htpasswd file, used for authenticating users
    against the Apache password file used for HTTP Basic Authentication.  This class
    validates a htpasswd file for existence.  It allows one to list the user entries
    found in a local htpasswd file, as well as validate a user against the htpasswd file.

    Parameters
    ----------
    path : str
        The path to a .htpasswd file.  Defaults to custom "htpass_path" config path or ~/.htpasswd
    """

    def __init__(self, path: str = None):
        path = path or cfg_params.get('htpass_path', None) or '~/.htpasswd'
        self.path = pathlib.Path(path).expanduser()

        self.htpass = None
        self._check_htpass()

    def __repr__(self) -> str:
        return f'<Htpass(path="{self.path}", valid={self.is_valid})>'

    @property
    def is_valid(self) -> bool:
        """ Checks if the htpass file exists and htpass object is loaded """
        return self.path.is_file() and self.htpass is not None

    def validate_user(self, username: str, password: str) -> bool:
        """ Validate a user entry in the htpass file

        Checks the provided username and password against the entry in
        the htpass file.

        Parameters
        ----------
        username : str
            A valid htpass username entry
        password : str
            The password for the provided username

        Returns
        -------
        bool
            True when the username/password validates successfully
        """
        if not self.is_valid:
            warnings.warn('htpass is not valid.  Cannot validate user.  Check file path.')
            return False
        return self.htpass.check_password(username, password)

    def _check_htpass(self) -> None:
        """ Check existence of the htpasswd file

        Reads in the htpass file with passlib.apache.HtpasswdFile

        Raises
        ------
        ImportError
            when passlib package not installed
        BrainError
            when the provided htpass file path does not exist
        """
        if not passlib:
            raise ImportError('passlib package not installed.  Cannot use Htpass.')

        if not self.path.is_file():
            raise BrainError(f'No .htpasswd file found at {self.path}!')

        self.htpass = HtpasswdFile(self.path)

    def list_users(self) -> list:
        """ Return a list of users in the htpasswd file

        Returns
        -------
        list
            list of users in the htpasswd file
        """
        if not self.is_valid:
            warnings.warn('htpass is not valid.  Cannot list users.  Check file path.')
            return None
        return self.htpass.users()



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: __init__.py
# Project: api
# Author: Brian Cherinka
# Created: Friday, 23rd October 2020 1:53:29 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Friday, 23rd October 2020 1:53:29 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

#
# Items to consider from the Brain
# - setting different API urls  (local, test, prod, public) - submit requests to development APIs
# - toggling different base API urls - submit requests to different APIs if available
# - tree replants for different base API urls - toggle tree replants for different APIs
# - wrapper to httpx or requests - provide convienences for submitting http requests
# - accessing API route mappings - look up available routes for a given API
# - JWT token or OAuth authentication - setting Auth precendence e.g. netrc over JWT
# - request/response compression schemes - toggle different compression algorithms

# - placeholder images - https://placeholder.com/
# - http tester - https://httpbin.org/




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: user.py
# Project: auth
# Author: Brian Cherinka
# Created: Thursday, 22nd October 2020 11:17:33 am
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Thursday, 22nd October 2020 11:17:33 am
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

from sdss_brain.auth import Netrc, Htpass
from sdss_brain.exceptions import BrainError
from sdss_brain.api.io import send_post_request

try:
    from collaboration.wiki import Credential
except ImportError:
    Credential = None


class User(object):
    """ Class representing a valid user

    This class represents a valid SDSS user.  It checks and validates the user
    against a local netrc file, a local htpasswd file, and against the SDSS
    collaboration.Credentials.

    Parameters
    ----------
    user : str
        the SDSS username

    Attributes
    ----------
    member : dict
        dictionary of member information for a validated user
    netrc : type[Netrc]
        A local Netrc instance, if any
    htpass : type[Htpass]
        A local Htpass instance, if any
    cred : type[Credential]
        A local SDSS Credential instance, if any
    """

    def __init__(self, user: str):
        self.user = user
        self.netrc = None
        self.htpass = None
        self.cred = None
        self.member = None

        self._valid_netrc = False
        self._valid_htpass = False
        self._valid_sdss_cred = False

        self._setup_auths()

    def __repr__(self) -> str:
        return (f'<User("{self.user}", netrc={self.is_netrc_valid}, htpass={self.is_htpass_valid}, '
                f'cred={self.is_sdss_cred_valid})>')

    def __str__(self):
        return self.user

    def _setup_auths(self):
        """ Setup the authenticators """
        # setup netrc
        try:
            self.netrc = Netrc()
        except BrainError:
            self.netrc = None
        else:
            self._valid_netrc = (self.user in self.netrc.read_netrc('data.sdss.org') or
                                 self.user in self.netrc.read_netrc('api.sdss.org'))

        # setup htpass
        try:
            self.htpass = Htpass()
        except BrainError:
            self.htpass = None
        else:
            self._valid_htpass = False

        # setup SDSS Credential
        if Credential:
            self.cred = Credential

    @property
    def is_netrc_valid(self) -> bool:
        """ Checks if the netrc credentials are validated for the given user """
        return False if not self.netrc else self._valid_netrc

    @property
    def is_htpass_valid(self) -> bool:
        """ Checks if htpasswd credentials are validated for the given user """
        return False if not self.htpass else self._valid_htpass

    @property
    def is_sdss_cred_valid(self) -> bool:
        """ Checks if SDSS credentials are validated for the given user """
        return self._valid_sdss_cred

    @property
    def validated(self) -> bool:
        """ Checks if user is validated """
        return any([self.is_netrc_valid, self.is_htpass_valid, self.is_sdss_cred_valid])

    @property
    def in_sdss(self) -> dict:
        """ Checks member status in SDSS-IV and SDSS-V """
        if self.member:
            return {'sdss4': self.member['sdss4']['has_sdss_access'],
                    'sdss5': self.member['sdss5']['has_sdss_access']}

    def validate_user(self, password: str = None) -> None:
        """ Validate the given user

        Validates the given user for netrc, htpass, and SDSS Credential
        authentication.  Attempts to extract a valid password from the netrc file
        for the given user.  Otherwise a password must be explicitly input.

        Parameters
        ----------
        password : str, optional
            The password for the given user, by default None

        Raises
        ------
        ValueError
            when the netrc username does not match the input user
        ValueError
            when no valid password is input nor extracted from the netrc
        """
        # look up user info from validated netrc
        if self.is_netrc_valid:
            try:
                user, passwd = self.netrc.read_netrc('data.sdss.org')
            except ValueError:
                user, passwd = self.netrc.read_netrc('api.sdss.org')

            if user != self.user:
                raise ValueError(f'netrc user {user} mismatched with input user {self.user}!')

            # if no password set, use the netrc password
            password = passwd if not password else password

        # ensure a password is specified
        if not password:
            raise ValueError('Must provide a password if none extracted from netrc.')

        # validate the htpasswd
        if self.htpass:
            self._valid_htpass = self.htpass.validate_user(self.user, password) or False

        # try to validate using SDSS Credentials
        if self.cred:
            cred = Credential(self.user, password)
            cred.authenticate_via_trac()
            self._valid_sdss_cred = cred.authenticated is True
        else:
            # TODO change this dev login to production site when ready
            cred_url = 'https://internal.sdss.org/dev/collaboration/api/login'
            data = send_post_request(cred_url, data={'username': self.user, 'password': password})
            self._valid_sdss_cred = data.get('authenticated', False) == 'True'
            self.member = data.get('member', None)




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: database.py
# Project: helpers
# Author: Brian Cherinka
# Created: Sunday, 18th October 2020 11:54:01 am
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Sunday, 18th October 2020 11:54:01 am
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import importlib
import inspect
from types import ModuleType
from typing import List, Union, Type

import sdssdb.peewee
import sdssdb.sqlalchemy
from sdssdb.connection import DatabaseConnection, PeeweeDatabaseConnection, SQLADatabaseConnection

from sdss_brain.exceptions import BrainError

db_type = Union[Type[PeeweeDatabaseConnection], Type[SQLADatabaseConnection],
                Type[sdssdb.peewee.BaseModel], Type[sdssdb.sqlalchemy.BaseModel],
                Type[ModuleType]]


class DatabaseHandler(object):
    """ Convenience class for handling an sdssdb database connection / ORM

    A convenience class to provide a simple API to any sdssdb database or ORM.
    Accepts as input any sdssdb peewee or sqlalchemy ORM Model, DatabaseConnection,
    or a sdssdb schema ORM python module. Given valid input will determine, which
    database, schema, set of ORMs, and primary ORM.

    Parameters
    ----------
    db_input : db_type
        The type of sdssdb database input.

    Attributes
    ----------
        orm : str
            Indicates whether the ORM is peewee or sqlalchemy
        model : `~sdssdb.peewee.BaseModel` | `~sdssdb.sqlalchemy.BaseModel`
            Any database ORM Model
        models : py module
            A database schema python module containing all ORM models
        db : `~sdssdb.connection.DatabaseConnection`
            A database connection object
        schema : str
            The schema name
        session : `~sqlalchemy.orm.session.Session`
            A sqlalchemy database session needed for querying
    """

    def __init__(self, db_input: db_type = None) -> None:
        self.orm = None
        self.model = None
        self.models = None
        self._session = None
        self.db = None
        self.schema = None

        self._determine_input(db_input)

    def __repr__(self) -> str:
        d = {'orm': self.orm, 'model': self.model.__name__ if self.model else None,
             'schema': self.schema, 'db': self.db.dbname if self.db else None}
        repr_items = ', '.join([m for m in map(lambda x: '{0}="{1}"'.format(
                     *x) if x[1] else '', d.items()) if m])
        return f'<DatabaseHandler ({repr_items})>'

    def _determine_input(self, value: db_type = None) -> None:
        """ Determine the type of database input

        Determines whether the input is a valid sdssdb ORM Model,
        DatabaseConnection, or schema module, and sets the instance
        attributes for orm, db, schema, model and models.

        Parameters
        ----------
        value : db_type, optional
            The type of ``sdssdb`` database input, by default None
        """
        if not value:
            return

        if any(self._is_a_model(value)):
            self._check_models(value)
        elif any(self._is_a_db(value)):
            self._check_dbs(value)
        elif any(self._is_a_schema(value)):
            self._check_schema(value)

    @staticmethod
    def _is_a_model(value: db_type) -> List[bool]:
        """ Conditional check if input is an ORM Model """
        if not inspect.isclass(value):
            return [None, None]

        pmodel, smodel = issubclass(value, sdssdb.peewee.BaseModel), issubclass(
            value, sdssdb.sqlalchemy.BaseModel)
        return [pmodel, smodel]

    def _check_models(self, value: db_type) -> None:
        """ Checks a sdssdb ORM

        Given any sdssdb peewee or sqlalchemy ORM, extracts
        relevant information on the ORM, database, schema name, and set of ORM
        models.

        Parameters
        ----------
        value : `~sdssdb.peewee.BaseModel` | `~sdssdb.sqlalchemy.BaseModel`
            A valid sdssdb peewee or sqla ORM
        """
        pmodel, smodel = self._is_a_model(value)

        self.model = value
        self.models = importlib.import_module(value.__module__)
        if pmodel:
            self.orm = 'peewee'
            self.schema = value._meta.schema
            self.db = value._meta.database
        elif smodel:
            self.orm = 'sqla'
            self.schema = value._schema
            self.db = getattr(self.models, 'database', None)
            self._set_session()

    @staticmethod
    def _is_a_db(value: db_type) -> List[bool]:
        """ Conditional check if input is a DatabaseConnection """
        pdatabase, sdatabase = isinstance(value, PeeweeDatabaseConnection), isinstance(
            value, SQLADatabaseConnection)
        return [pdatabase, sdatabase]

    def _check_dbs(self, value: Type[DatabaseConnection]) -> None:
        """ Checks a sdssdb database connection

        Given a sdssdb peewee or sqlalchemy database connection, extracts
        relevant information on the ORM, database, schema name, and set of ORM
        models.

        Parameters
        ----------
        value : `~sdssdb.connection.DatabaseConnection`
            A valid sdssdb database connection
        """
        pdatabase, sdatabase = self._is_a_db(value)
        self.db = value
        if pdatabase:
            self.orm = 'peewee'
        elif sdatabase:
            self.orm = 'sqla'
            self._set_session()

    @staticmethod
    def _is_a_schema(value: db_type) -> List[bool]:
        """ Conditional check if input is a schema module """
        is_module = inspect.ismodule(value)
        name = value.__name__ if is_module else ''
        pschema, sschema = is_module and 'peewee' in name, is_module and 'sqlalchemy' in name
        return [pschema, sschema]

    def _check_schema(self, value: Type[ModuleType]) -> None:
        """ Checks a sdssdb schema module file

        Given a schema module files, extracts relevant information
        on the ORM, database, schema name, and set of ORM models

        Parameters
        ----------
        value : py module
            A valid sdssdb database schema module file
        """
        pschema, sschema = self._is_a_schema(value)

        self.models = value
        self.db = getattr(value, 'database', None)
        self.schema = self.models.__name__.split('.', 2)[-1]

        if pschema:
            self.orm = 'peewee'
        elif sschema:
            self.orm = 'sqla'
            self._set_session()

    @property
    def connected(self) -> bool:
        """ Returns True if the database is connected """
        return self.db.connected if self.db else None

    def _set_session(self):
        """ Sets the sqla database Session if db is connected """
        if self.connected:
            self._session = self.db.Session()

    @property
    def session(self):
        """ A sqlalchemy database session needed for querying

        Returns
        -------
        `~sqlalchemy.orm.session.Session`
            A sqlalchemy database session needed for querying

        Raises
        ------
        AttributeError
            when the orm is "peewee"
        """

        if self.orm == 'peewee':
            raise AttributeError('peewee ORMs do not use sessions. To query, access the model '
                                 'directly or load a sqlalchemy ORM')
        return self._session

    def load_schema(self, schema: str):
        """ Load an sdssdb schema module

        Loads an input sdssdb schema module name from the currently loaded database
        and into the handler. The input name can be dot-qualified as "[orm].[database].[schema]"
        and it will attempt to find and load the correct module.  For example
        ``targetdb``, ``sdss5db.targetdb``, ``peewee.sdss5db.targetdb`` are all
        valid names.

        Parameters
        ----------
        schema : str
            The name of an sdssdb schema module

        Raises
        ------
        TypeError
            when the input is not a string
        BrainError
            when no database connection is present
        BrainError
            when no schema module is found by importlib.import_module
        """

        if type(schema) != str:
            raise TypeError(f'Input {schema} must be a string.')

        if not self.db:
            raise BrainError(f'No db present. Cannot load schema {schema}.')

        orm = 'sqlalchemy' if self.orm == 'sqla' else 'peewee'
        dbname = self.db.dbname if self.db else ''
        if schema.count('.') == 0:
            modname = f'sdssdb.{orm}.{dbname}.{schema}'
        elif schema.count('.') == 1:
            modname = f'sdssdb.{orm}.{schema}'
        elif schema.count('.') == 2:
            modname = f'sdssdb.{schema}'
        else:
            modname = schema

        try:
            self.models = importlib.import_module(modname)
        except ModuleNotFoundError as e:
            raise BrainError(f'No module found matching {modname}') from e
        else:
            self.schema = self.models.__name__.split('.', 2)[-1]
            self.model = None

    def load_model(self, model: str):
        """ Loads an ORM model

        Loads an input ORM model name from the currently loaded database
        and schema into the handler.

        Parameters
        ----------
        model : str
            The name of an ORM model to load

        Raises
        ------
        BrainError
            when no valid schema is set
        AttributeError
            when no model is found within the loaded schema
        """

        if not self.models:
            raise BrainError('No valid schema set containing ORM models.')

        model_obj = getattr(self.models, model, None)
        if not model_obj:
            raise AttributeError(f'schema {self.schema} does not have model {model}')

        self.model = model_obj

    def close(self):
        """ Close database connections and sessions """
        if self.orm == 'sqlalchemy':
            self.session.close()
        elif self.orm == 'peewee':
            self.db.close()




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: io.py
# Project: helpers
# Author: Brian Cherinka
# Created: Wednesday, 7th October 2020 10:54:12 am
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Wednesday, 7th October 2020 10:54:12 am
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import pathlib
import gzip
from typing import Union
from astropy.io import fits
from sdss_brain import log
from sdss_brain.api.client import SDSSClient
from sdss_brain.config import config
from sdss_brain.exceptions import BrainError


def get_mapped_version(name: str, release: str = None, key: str = None) -> Union[dict, str]:
    ''' Get a version id mapped to a release number

    For a given named category, looks up the "mapped_versions" attribute from
    the configuration yaml file and returns a version number that has been mapped
    to a specific release. For example, for manga, DR16 maps to drpver='v2_4_3'
    and dapver='2.2.1'. This can be useful when needing to specify certain versions
    when defining paths to files.

    Parameters
    ----------
        name : str
            The name of the set of versions to access
        release : str
            The SDSS release.  Default is config.release.
        key : str
            Optional name of dictionary key to access specific value

    Returns
    -------
        version : dict|str
            A version number corresponding to a given release

    Example
    -------
        >>> # access the MaNGA versions for release DR16
        >>> get_mapped_version('manga', release='DR16')
            {'drpver': 'v2_4_3', 'dapver': '2.2.1'}

        >>> # access specific key
        >>> get_mapped_version('manga', release='DR16', key='drpver')
            'v2_4_3'
    '''
    # if release is a work release, return nothing
    if release.lower() == 'work':
        return None

    # get the mapped_versions attribute from the configuration
    mapped_versions = config._custom_config.get('mapped_versions', None)
    assert mapped_versions, 'mapped_versions must be defined'
    if type(mapped_versions) != dict:
        raise TypeError('mapped_versions must be a dictionary')

    # ensure that the name is a valid entry
    if name not in mapped_versions:
        raise ValueError(f'{name} not found in mapped_versions dictionary')

    versions = mapped_versions.get(name, None)
    if type(versions) != dict:
        raise TypeError(f'release versions for {name} must be a dictionary')

    # ensure that the release is a valid entry
    release = release or config.release
    version = versions.get(release, None)
    if not version:
        raise ValueError(f'no mapped_version found for release {release} in {name}. '
                         'Check the sdss_brain.yml config file.')

    # check for a specific key in the version dictionary
    if key and type(version) == dict:
        version = version.get(key, None)

    return version


def load_fits_file(filename: str) -> fits.HDUList:
    ''' Load a FITS file

    Opens and loads a FITS file with astropy.io.fits.

    Parameters
    ----------
        filename : str
            A FITS filen to open

    Returns
    -------
        hdulist : `~astropy.io.fits.HDUList`
            an Astropy HDUList
    '''

    path = pathlib.Path(filename)
    if not path.exists() and path.is_file():
        raise FileNotFoundError('input filename must exist and be a file')

    assert '.fits' in path.suffixes, 'filename is not a valid FITS file'

    try:
        hdulist = fits.open(path)
    except (IOError, OSError) as err:
        log.error(f'Cannot open FITS file {filename}: {err}')
        raise BrainError(f'Failed to open FITS files {filename}: {err}') from err
    else:
        return hdulist


def load_from_url(url: str, no_progress: bool = None) -> fits.HDUList:
    ''' Load a file from a remote url using a get request

    Streams url content with httpx.stream and pipes the response contents
    into an Astropy FITS file.

    Parameters
    ----------
        url : str
            A url path to a filename
        no_progress : bool
            If True, turns off the tqdm progress bar

    Returns
    -------
        an Astropy `~astropy.io.fits.HDUList`
    '''
    with SDSSClient(url, no_progress=no_progress) as client:
        client.request(method='stream')
        if url.endswith('.gz'):
            return fits.open(gzip.open(client.data, 'rb'))
        else:
            return fits.open(client.data)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: handler.py
# Project: api
# Author: Brian Cherinka
# Created: Friday, 30th October 2020 4:16:04 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Friday, 30th October 2020 4:16:04 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import os
import re
from typing import Type, Union
from sdss_brain.config import config
from sdss_brain.api.client import SDSSClient, SDSSAsyncClient
from sdss_brain.api.manager import apim, ApiProfile, strjoin
from sdss_brain.exceptions import BrainError


api_type = Union[str, list, tuple, Type[ApiProfile]]


class ApiHandler(object):
    """ Convenience class for handling an API profile, and SDSS remote http client

    A convenience class to provide a simple API to the client for sending remote http requests
    using the ``SDSSClient``, with access to the underlying SDSS API profile.
    Accepts as input any string URL, API profile name or instance of ``ApiProfile``, as well as
    a tuple or list of valid "(api profile name, url route segment, optional domain name)", e.g.
    "('marvin', '/general/getroutemap').  Given valid input will attempt to determine, which URL,
    API profile, and SDSS remote client.

    Parameters
    ----------
    api_input : api_type, optional
        The type of API input, by default None
    async_client : bool, optional
        If True, instructs to use the `~sdss_brain.api.client.SDSSAsyncClient`, by default False

    Attributes
    ----------
    url : str
        The url or route segment
    api : `~sdss_brain.api.manager.ApiProfile`
        The SDSS API profile instance
    client: `~sdss_brain.api.client.SDSSClient`
        The client for sending http requests
    """

    def __init__(self, api_input: api_type = None, async_client: bool = False, release: str = None):
        self.api = None
        self.url = None
        self.client = None
        self.timeout = 30
        self.release = release or config.release
        self._kls = SDSSAsyncClient if async_client else SDSSClient

        self._determine_input(api_input)

    def __repr__(self) -> str:
        return f'<ApiHandler(api="{str(self.api)}", url="{self.url}")>'

    def _determine_input(self, value: api_type = None) -> None:
        """ Determine the type of API input

        Determine the type of API input.  Valid input can be either a valid
        URL string or API profile name, a tuple of (API profile name, route segment)
        or a valid SDSS ApiProfile instance.  Attempts to set the ``url``, ``api``, and
        ``client`` attributes.

        Parameters
        ----------
        value : api_type, optional
            The type of API input, by default None.
        """
        # if nothing, do nothing
        if not value:
            return

        if type(value) == str:
            self._check_url(value)
        elif isinstance(value, (list, tuple)):
            self._check_api(value)
        elif isinstance(value, ApiProfile):
            self._check_profile(value)

    def _check_url(self, value: str) -> None:
        """ Check a URL string or API name

        Check if the input string is a URL starting with http
        or the name of an available SDSS API profile.  If able, sets
        a valid URL, API, and SDSS remote client.

        Parameters
        ----------
        value : str
            the input URL string or API profile name
        """
        # check if the input is a URL starting with http
        if value.startswith('http'):
            self.url = value
            # extract the API and domain names and set the API accordingly
            api, domain = apim.identify_api_from_url(value)
            if api:
                self.api = apim.apis[api]
                if domain:
                    self.api.change_domain(domain)
        else:
            # set the API profile
            self.api = apim.apis.get(value, None)

        # set the SDSS client
        self.client = self._kls(self.url, use_api=self.api, timeout=self.timeout,
                                release=self.release)
        if self.api and not self.client.api:
            self.client.set_api(self.api)

    def _check_api(self, value: Union[tuple, list]) -> None:
        """ Check the API/route tuple

        Check the input tuple or list for a valid API profile name and
        URL route.  Can optionally include a domain name as the third element in the list.
        Input syntax must be "(api name, URL route, domain name)".  If able, sets
        a valid API, the route segment as URL, and SDSS remote client.

        Parameters
        ----------
        value : Union[tuple, list]
            A valid API name and URL route segment.  Optionally include a domain name.

        Raises
        ------
        ValueError
            when the input list is not 2+ elements in length
        ValueError
            when the domain name is not in the available domain list
        """
        if len(value) < 2:
            raise ValueError('Input tuple/list must have at least two parts.')

        name, route, *extra = value
        self.api = apim.apis.get(name, None)
        if extra:
            domain = extra[0]
            if domain not in apim.domains:
                raise ValueError(f'Domain {domain} is not an available SDSS domain.')
            else:
                self.api.change_domain(domain)
        self.url = route
        self.client = self._kls(self.url, use_api=self.api, timeout=self.timeout,
                                release=self.release)

    def _check_profile(self, value: Type[ApiProfile]) -> None:
        """ Check the API Profile

        Check the input API Profile, sets it as the api and
        sets the SDSS remote client.

        Parameters
        ----------
        value : Type[ApiProfile]
            A valid SDSS ApiProfile
        """
        self.api = value
        self.client = self._kls(use_api=self.api, timeout=self.timeout, release=self.release)

    def extend_url(self, route: str) -> str:
        """ Extend the current url by a route segment

        Extend the current url by the given route segment and
        return a temporary new url.  Meant to easily extend one base
        route to a deeper sub-route.

        Parameters
        ----------
        route : str
            A url route segment

        Returns
        -------
        str
            A new combined url
        """
        if not self.url:
            raise AttributeError('No preset base url.  Cannot extend it. Consider setting one.')

        return strjoin(self.url, route)

    def resolve_url(self, params: dict):
        """ Resolve any url bracket parameters

        Format the url to replace any bracket name arguments with the
        input parameters.  All bracket arguments must be replaced and filled
        in before a valid http request can be sent.

        Parameters
        ----------
        params : dict
            A dictionary of parameters to format the url with

        Raises
        ------
        ValueError
            when no url is set
        TypeError
            when input params is not a dictionary
        """
        if not self.url:
            raise ValueError('No url to resolve.  Consider setting one.')

        if type(params) != dict:
            raise TypeError('Input params must be a dictionary.')

        self.url = self.url.format(**params)
        self.client.set_url(self.url)

    @property
    def has_valid_url(self) -> bool:
        """ Check if the url has no bracket arguments """
        if self.url:
            parts = self.extract_url_brackets()
            return not parts

    def extract_url_brackets(self) -> list:
        """ Extract named parameters from any url brackets

        Extracts any named parameter arguments from a url, for cases where brackets
        ({}) are used to denote an argument subtitution.  For example,
        "cubes/{plateifu}/extensions/{extname}/{x}/{y}"

        Returns
        -------
        list
            A list of url parameter names
        """
        parts = []
        if self.url:
            parts = re.findall(r'{(.*?)}', self.url)
        return parts

    def close(self):
        """ Close the synchronous SDSSClient httpx client """
        try:
            self.client.client.close()
        except AttributeError:
            pass

    def load_url(self, route: str) -> None:
        """ Loads a url constructed from an API base url

        Construct a new url, given the input route segment, for a given
        API profile.

        Parameters
        ----------
        route : str
            The url path route segment

        Raises
        ------
        TypeError
            when input route is not a string
        BrainError
            when no API profile is set on the handler
        """
        if type(route) != str:
            raise TypeError('Input route must be a string.')

        if not self.api:
            raise BrainError('No API profile is set.  Cannot construct a url.')

        self.url = self.api.construct_route(route)
        self.client.set_url(self.url)

    def change_domain(self, domain: str, test: bool = None) -> None:
        """ Change the domain of the set API

        Changes the domain on the loaded API and updates the http client
        to use the new API.

        Parameters
        ----------
        domain : str
            The name of the domain to switch to
        test : bool, optional
            If True, uses the test server, by default None

        Raises
        ------
        AttributeError
            when no API profile is set on the handler
        """
        if not self.api:
            raise AttributeError('Cannot change domain.  No api profile set.')

        # change to the new API domain and optionally, the test server
        self.api.change_domain(domain)
        if test:
            self.api.change_path(test=test)

        # update the API on the client
        self.client.set_api(self.api)

        # reset the url
        self.client.set_url(self.url)





# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: core.py
# Project: sdss_brain
# Author: Brian Cherinka
# Created: Sunday, 15th March 2020 4:53:35 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Wednesday, 18th March 2020 4:08:26 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import abc
import warnings
from sdss_brain.config import config
from sdss_brain.exceptions import BrainError
from sdss_brain.helpers import db_type
from sdss_brain.api.handler import api_type, ApiHandler
from sdss_brain.mixins.mma import MMAccess, MMAMixIn
from astropy.io import fits


class Base(abc.ABC):
    ''' abstract base class for tools '''

    def __new__(cls, *args, **kwargs):
        # set the correct MMA class
        if MMAccess in cls.mro():
            cls._mma = MMAccess
        else:
            cls._mma = MMAMixIn

        return super().__new__(cls)

    @abc.abstractmethod
    def _load_object_from_file(self) -> None:
        pass

    @abc.abstractmethod
    def _load_object_from_db(self) -> None:
        pass

    @abc.abstractmethod
    def _load_object_from_api(self) -> None:
        pass


class HindBrain(Base):
    ''' Base class for utilizing the MMA mixin

    This is a convenience class with the `~sdss_brain.mixins.mma.MMAccess` already implemented.
    This class initializes the ``MMAccess`` and provides logic to load data based
    on the data_origin.  It also provides a simple ``repr``.

    In addition to any abstractmethod from the MMA, this class contains three abstractmethods you
    must override when subclassing.

    - **_load_object_from_file**: defines data load/handling from a local file
    - **_load_object_from_db**: defines data load/handling from a local database
    - **_load_object_from_api**: defines data load/handling from a remote API

    Parameters
    ----------
        data_input : str
            The file or name of target data to load
        filename : str
            The absolute filepath to data to load
        objectid : str
            The object identifier of the data to load
        mode : str
            The operating mode: auto, local, or remote
        release : str
            The data release of the object, e.g. "DR16"
        download : bool
            If True, downloads the object locally with sdss_access
        ignore_db : bool
            If True, ignores any database connection for local access
        use_db : db_type, see `~sdss_brain.helpers.database.DatabaseHandler`
            a database ORM or connection to override the default with
        use_api : api_type, see `~sdss_brain.api.handler.ApiHandler`
            An API or URL route to use for remote access to the object
        async_client : bool
            If True, sets the remote http client to `~sdss_brain.api.client.SDSSAsyncClient`
    Attributes
    ----------
        db : `~sdss_brain.helpers.database.DatabaseHandler`
            A db handler for any loaded sdssdb ORM or db connection for the object
        mapped_version : str
            The name of survey/category in the mapped_versions dictionary
        remote : `~sdss_brain.api.handler.ApiHandler`
            An api handler for loaded APIs or urls, and sending http requests
    '''
    _db = None
    mapped_version = None
    data_origin = None
    _api = None
    async_client = False

    def __new__(cls, *args, **kwargs):
        # set any work versions
        cls.set_work_version(config.work_versions)
        return super().__new__(cls)

    def __init__(self, data_input: str = None, filename: str = None,
                 objectid: str = None, mode: str = None,
                 release: str = None, download: bool = None,
                 ignore_db: bool = None, use_db: db_type = None, version: str = None,
                 use_api: api_type = None, async_client: bool = None) -> None:

        # set a version for sdsswork data
        checked_release = release or config.release
        if version:
            self.set_work_version(version)
            if checked_release.lower() != 'work':
                raise BrainError('version is only used for "work" data. '
                                 'Please set the input or config release to "WORK"')
        else:
            if not self._version and checked_release.lower() == 'work':
                raise BrainError('You are using a "work" release but have no work versions set! '
                                 'Try setting a global "work_version" dict or specify a "version" '
                                 'input!')

        # initialize the MMA
        self._mma.__init__(self, data_input=data_input, filename=filename,
                           objectid=objectid, mode=mode,
                           release=release, download=download,
                           ignore_db=ignore_db, use_db=use_db or self._db)

        # set up any ApiHandler
        remote = use_api or self._api or config.apis.profile
        async_client = async_client or self.async_client
        if remote:
            self._api = ApiHandler(remote, async_client=async_client)
            self._check_remote_url()

        # load the data
        if self.data_origin == 'file':
            self._load_object_from_file()
        elif self.data_origin == 'db':
            self._load_object_from_db()
        elif self.data_origin == 'api':
            self._load_object_from_api()

    def __repr__(self):
        objname = f"objectid='{self.objectid}'" if self.objectid else f"filename='{self.filename}'"
        return (f"<{self.__class__.__name__} {objname}, mode='{self.mode}', "
                f"data_origin='{self.data_origin}'>")

    @property
    def remote(self):
        """ An instance of an ApiHandler """
        return self._api

    def resolve_remote_url(self):
        """ Resolve the remote url using instance attributes

        Attempts to resolve the remote url and match the named arguments
        in the url to any instance attributes that are set.  Otherwise issues a warning
        for the user to resolve manually.
        """
        if not self.remote or (self.remote and not self.remote.url):
            return

        names = self.remote.extract_url_brackets()
        if names:
            params = {}
            for name in names:
                if hasattr(self, name):
                    params[name] = getattr(self, name)
            try:
                self.remote.resolve_url(params)
            except KeyError as exc:
                raise BrainError(f'Cannot resolve remote url. Missing keys {exc} found.')from exc
            else:
                if not self.remote.has_valid_url:
                    raise BrainError('Still invalid url')

    def _check_remote_url(self):
        """ Check for a valid remote url and issue warning otherwise """
        if self._api and self._api.has_valid_url is False:
            try:
                self.resolve_remote_url()
            except BrainError as exc:
                warnings.warn('API url contains bracket arguments. Http request will not send '
                              'correctly.  Consider resolving the url with '
                              f'self.remote.resolve_url. {exc}')

    def __del__(self):
        ''' Destructor for closing open objects '''
        self._close()

    def _close(self):
        ''' close open object for each data_origin '''
        # close open FITS files
        if self.data_origin == 'file' and isinstance(self.data, fits.HDUList):
            self.data.close()
        elif self.data_origin == 'db':
            if hasattr(self, 'db') and self.db is not None:
                self.db.close()
            self.data = None
        elif self.data_origin == 'api':
            if hasattr(self, 'remote') and self.remote is not None:
                self.remote.close()

    def __enter__(self):
        ''' constructor for context manager '''
        return self

    def __exit__(self, type, value, traceback):
        ''' destructor for context manager '''
        self._close()
        return True

    @classmethod
    def set_work_version(cls, value: dict) -> None:
        """ Set the work version for the given class

        Sets the versions used by sdswork on the given class.  This takes
        precendence over any versions set in the global config. Input is a valid
        dictionary containing version names and numbers as key value pairs, e.g.
        `{'drpver':'v2_4_3', 'run2d':'v1_1_1', 'apred':'r12'}`.

        The input dictionary is merged with any values specified on the config class.

        Parameters
        ----------
        values : dict, optional
            Dictionary of version names:numbers needed in paths, by default {}

        Raises
        ------
        ValueError
            when input value is not a dictionary
        """
        if type(value) != dict:
            raise ValueError(f'input verion must be a dictionary')

        # update any existing work versions with the new values
        wv = config.work_versions.copy()
        wv.update(value)
        cls._version = wv

    @classmethod
    def set_database_object(cls, value: db_type) -> None:
        """ Sets up the MMA to create a new db handler

        Sets up the MMA to create a new `~sdss_brain.helpers.database.DatabaseHandler`
        on the given class.  Valid input can be any sdssdb ORM Model,
        DatabaseConnection, or schema module.

        Parameters
        ----------
        value : db_type
            The type of ``sdssdb`` database input
        """
        # set this straight to the value since the MMA converts it into a DatabaseHandler
        cls._db = value

    @classmethod
    def set_api_object(cls, value: api_type) -> None:
        """ Sets up the tool to create a new API handler

        Sets up the tool to create a new `~sdss_brain.api.handler.ApiHandler` on the
        given class.  Valid input can be any explicit url, a string API profile name,
        or a tuple containing (API profile name, url route segment, optional domain name).

        Parameters
        ----------
        value : api_type
            The type of remote API input
        """
        cls._api = value


class Brain(HindBrain, MMAccess):
    """ The hind Brain with support for ``sdss_access``

    See `~HindBrain`, `~sdss_brain.mixins.mma.MMAccess`, and `~sdss_brain.mixins.access.AccessMixIn`
    for detailed information.

    """


class BrainNoAccess(HindBrain, MMAMixIn):
    """ A version of `~Brain` without support for ``sdss_access``

    See `~HindBrain` and `~sdss_brain.mixins.mma.MMAMixIn` for detailed
    information.

    """



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: mma.py
# Project: mixins
# Author: Brian Cherinka
# Created: Thursday, 8th October 2020 11:23:28 am
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Thursday, 8th October 2020 11:23:28 am
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import abc
import six
import pathlib
import os
import warnings

from sdss_brain import log
from sdss_brain.mixins.access import AccessMixIn
from sdss_brain.config import config
from sdss_brain.exceptions import BrainError
from sdss_brain.helpers import DatabaseHandler, db_type


__all__ = ['MMAMixIn', 'MMAccess']


class MMAMixIn(abc.ABC):
    ''' Mixin for implementing multi-modal data access

    This is a mixin class that adds multi-modal data access to any class
    that subclasses from this one.  The MMA allows toggling between local
    and remote data access modes, or leaving it on automatic.  Local mode
    access tries to load data via a database, if one exists, otherwise it loads
    data via a local filepath.  Remote mode will try to load data over an API.
    When the mode is set to "auto", it automatically tries to first load things
    locally, and then remotely.  Depending on the mode and logic, the MMA will
    set data_origin to either `file`, `db`, or `api`.

    Note that this class does not provide the logic for loading data from a db, over an API,
    or from a file.  The user must provide that logic in a subclass.

    This mixin contains three abstractmethods you must override when subclassing.

    - **_parse_inputs**: provides logic to parse ``data_input`` into either filename or objectid
    - **download**: a method for downloading a data file to a local disk
    - **get_full_path**: a method for generating the absolute file path on disk to a file

    Parameters
    ----------
        data_input : str
            The file or name of target data to load
        filename : str
            The absolute filepath to data to load
        objectid : str
            The object identifier of the data to load
        mode : str
            The operating mode: auto, local, or remote
        release : str
            The data release of the object, e.g. "DR16"
        download : bool
            If True, downloads the object locally with sdss_access
        ignore_db : bool
            If True, ignores any database connection for local access
        use_db : db_type, see `~sdss_brain.helpers.database.DatabaseHandler`
            a database ORM or connection to override the default with

    Attributes
    ----------
        release : str
            The current data release loaded

    '''

    def __init__(self, data_input: str = None, filename: str = None, objectid: str = None,
                 mode: str = None, release: str = None, download: bool = None, ignore_db:
                 bool = False, use_db: db_type = None):
        # data attributes
        self._db = DatabaseHandler(use_db) if use_db else None
        self.filename = filename
        self.objectid = objectid
        self.data_origin = None

        # inputs or config variables
        self.mode = mode or config.mode
        self._release = release or config.release
        self._forcedownload = download or config.download
        self._ignore_db = ignore_db or config.ignore_db

        # determine the input
        self._determine_inputs(data_input)

        assert self.mode in ['auto', 'local', 'remote'], 'mode must be auto, local, or remote'
        assert self.filename is not None or self.objectid is not None, 'no inputs set.'

        # perform the multi-modal data access
        if self.mode == 'local':
            self._do_local()
        elif self.mode == 'remote':
            self._do_remote()
        elif self.mode == 'auto':
            try:
                self._do_local()
            except BrainError as ee:

                if self.filename:
                    # If the input contains a filename we don't want to go into remote mode.
                    raise(ee)
                else:
                    log.debug('local mode failed. Trying remote now.')
                    self._do_remote()

        # Sanity check to make sure data_origin has been properly set.
        assert self.data_origin in ['file', 'db', 'api'], 'data_origin is not properly set.'

    @property
    def release(self) -> str:
        """ Returns the release. """

        return self._release

    @release.setter
    def release(self, value: str):
        """Fails when trying to set the release after instantiation."""
        raise BrainError('the release cannot be changed once the object has been instantiated.')

    @property
    def db(self):
        """ an instance of a DatabaseHandler """
        return self._db

    def _do_local(self) -> None:
        """ Check if it's possible to load the data locally."""

        if self.filename:

            # check if the file exists locally
            if self.filename.exists():
                self.mode = 'local'
                self.data_origin = 'file'
            else:
                raise BrainError('input file {0} not found'.format(self.filename))

        elif self.objectid:

            # prioritize a database unless explicitly set to ignore
            if self.db and self.db.connected and not self._ignore_db:
                self.mode = 'local'
                self.data_origin = 'db'
            else:

                # retrieve the full local sdss_access path
                fullpath = self.get_full_path()

                if fullpath and os.path.exists(fullpath):
                    self.mode = 'local'
                    self.filename = pathlib.Path(fullpath)
                    self.data_origin = 'file'
                else:
                    # optionally download the file
                    if self._forcedownload:
                        self.download()
                        self.data_origin = 'file'
                    else:
                        raise BrainError('failed to retrieve data using '
                                         'input parameters.')

    def _do_remote(self) -> None:
        """ Check if remote connection is possible."""

        if self.filename:
            raise BrainError('filename not allowed in remote mode.')
        else:
            self.mode = 'remote'
            self.data_origin = 'api'

    def _determine_inputs(self, data_input: str) -> None:
        """ Determines what inputs to use in the decision tree.

        Parameters
        ----------
            data_input : str
                The input string to attempt to parse into a filename or object id
        """

        parsed_input = None

        if data_input:
            assert self.filename is None and self.objectid is None, \
                'if input is set, filename and objectid cannot be set.'

            if not isinstance(data_input, (six.string_types, pathlib.Path)):
                raise TypeError('input must be a string or pathlib.Path')

            # parse the input data into either a filename or objectid
            parsed_input = self._parse_input(data_input)
            if not parsed_input:
                self.filename = data_input
            else:
                assert isinstance(
                    parsed_input, dict), 'return value of _parse_input must be a dict'
                self.filename = parsed_input.get('filename', None)
                self.objectid = parsed_input.get('objectid', None)

        # ensure either filename or objectid is specified
        if self.filename is None and self.objectid is None:
            raise BrainError('no inputs defined. filename and objectid are both None')

        # convert filename to a pathlib.Path and resolve a relative name
        # not using pathlib.resolve to preserve symlinks
        if self.filename:
            self.filename = pathlib.Path(os.path.abspath(self.filename))
            # issue a warning if the release is not indicated in the filename; possible mismatch
            if self.release.lower() not in self.filename.as_posix():
                warnings.warn('Your filename may not match the release indicated.  Path parameters '
                              'may not be extracted properly.  Try setting the release to match the '
                              'known file version.')

        # attempt to update the access path parameters from the filename or parsed data input
        self._update_access_params(params=parsed_input)

        # check for any misaligments and misassignments
        if self.filename:
            self.objectid = None

            if self.mode == 'remote':
                raise BrainError('filename not allowed in remote mode.')

            assert self.filename.exists, \
                'filename {} does not exist.'.format(str(self.filename))

        elif self.objectid:
            assert not self.filename, 'invalid set of inputs.'

    @abc.abstractmethod
    def _parse_input(self, value: str) -> dict:
        ''' Parses the input value to determine the kind of input

        This method must be overridden by each subclass and contains the logic
        to determine the kind of input passed into it, i.e. either a filename or an
        object identification string.  This method accepts a single argument which is the
        string `data_input` and must return a dictionary containing at least keys
        for "filename" and "objectid".

        Parameters
        ----------
            value : str
                The data_input string to attempt to parse
        '''

    @abc.abstractmethod
    def download(self) -> None:
        ''' Abstract method to download a file '''
        pass

    @abc.abstractmethod
    def get_full_path(self) -> str:
        ''' Abstract method to return a full local file path '''
        pass

    @property
    def is_access_mixedin(self) -> bool:
        ''' Checks if the `~sdss_brain.mixins.access.AccessMixIn` is included '''
        return hasattr(self, 'path_name') and hasattr(self, 'access')

    def _update_access_params(self, params: str = None) -> None:
        ''' Updates the path_params attribute with extracted parameters

        Parameters
        ----------
            params : dict
                The output from the _parse_input method
        '''
        if self.is_access_mixedin and self.path_name:
            if self.filename:
                # attempt to extract the path_params from the filename
                params = self.access.extract(self.path_name, self.filename)
                if params:
                    self._setup_access(params, origin='file')
            elif self.objectid:
                # set attributes from extracted parse_input
                self._set_parsed_attributes(params)
                # run the set_access_path_params method
                self._set_access_path_params()
                # set attributes from the path_params
                self._setup_access(self.path_params, origin='object')
        elif not self.is_access_mixedin and params:
            # for non-access, set attributes from extracted parse_input
            self._set_parsed_attributes(params)

    def _set_parsed_attributes(self, params: dict) -> None:
        """ Set instance attributes from any extracted params from parse_input

        Parameters
        ----------
            params : dict
                The output from the _parse_input method
        """
        if not params:
            return

        if type(params) != dict:
            raise TypeError('the output of parse_input must be a dictionary')
        for key, val in params.items():
            if key in ['filename', 'objectid', 'parsed_groups', 'parsed_inputs']:
                continue

            # skip if a class attribute already exists
            if hasattr(self.__class__, key):
                continue

            # if a work version is set for the given key, and no value is set, then
            # use the work version
            work_ver = self._version.get(key, None)
            val = work_ver if work_ver and val is None else val

            setattr(self, key, val)


class MMAccess(AccessMixIn, MMAMixIn):
    """ Class that mixes in the sdss_access functionality with the MMA

    This is a mixin class that adds multi-modal data access to any class
    that subclasses from this one.  The MMA allows toggling between local
    and remote data access modes, or leaving it on automatic.  Local mode
    access tries to load data via a database, if one exists, otherwise it loads
    data via a local filepath.  Remote mode will try to load data over an API.
    When the mode is set to "auto", it automatically tries to first load things
    locally, and then remotely.  Depending on the mode and logic, the MMA will
    set data_origin to either `file`, `db`, or `api`.

    Note that this class does not provide the logic for loading data from a db, over an API,
    or from a file.  The user must provide that logic in a subclass.

    This mixin contains two abstractmethods you must override when subclassing.

    - **_set_access_path_params**: sets the arguments needed by `sdss_access`
    - **_parse_inputs**: provides logic to parse ``data_input`` into either filename or objectid

    Parameters
    ----------
        data_input : str
            The file or name of target data to load
        filename : str
            The absolute filepath to data to load
        objectid : str
            The object identifier of the data to load
        mode : str
            The operating mode: auto, local, or remote
        release : str
            The data release of the object, e.g. "DR16"
        download : bool
            If True, downloads the object locally with sdss_access
        ignore_db : bool
            If True, ignores any database connection for local access
        use_db : `~sdssdb.connection.DatabaseConnection`
            a database connection to override the default with

    Attributes
    ----------
        release : str
            The current data release loaded
        access : `~sdss_access.sync.Access`
            An instance of ``sdss_access`` using for all path creation and file downloads

    """
    pass



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: io.py
# Project: api
# Author: Brian Cherinka
# Created: Friday, 30th October 2020 1:22:42 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Friday, 30th October 2020 1:22:43 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import httpx
from sdss_brain.exceptions import BrainError


def send_post_request(url: str, data: dict = None) -> dict:
    """ A simple httpx post request

    A simple standalone httpx post request to a specified url, and
    given an optional data payload.

    Parameters
    ----------
    url : str
        The url to send the request to
    data : dict, optional
        Input data to send along with the request, by default None

    Returns
    -------
    dict
        Extracted response data from response.json()

    Raises
    ------
    BrainError
        when an error occurs sending the request
    """
    try:
        resp = httpx.post(url, data=data)
    except httpx.RequestError as exc:
        raise BrainError(f'An error occurred requesting {exc.request.url!r}') from exc
    else:
        resp.raise_for_status()
        data = resp.json()
    return data



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: netrc.py
# Project: auth
# Author: Brian Cherinka
# Created: Wednesday, 21st October 2020 2:52:15 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Wednesday, 21st October 2020 2:52:15 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import netrc
import pathlib
import warnings
from sdss_brain import cfg_params
from sdss_brain.exceptions import BrainError


class Netrc(object):
    """ Class representing a netrc file

    This class represents a local .netrc file, used for authenticating users
    against remote machines. This class validates a netrc file for existence and
    correct permissions.  It also checks its hosts against the allowed
    SDSS host domains data.sdss.org and api.sdss.org and ensures their presence.

    Parameters
    ----------
    path : str
        The path to a .netrc file.  Defaults to custom "netrc_path" config path or ~/.netrc
    """

    def __init__(self, path: str = None):
        path = path or cfg_params.get('netrc_path', None) or '~/.netrc'
        self.path = pathlib.Path(path).expanduser()
        self.allowed_hosts = ['data.sdss.org', 'api.sdss.org', 'magrathea.sdss.org']
        self._valid_hosts = {}

        self.check_netrc()

    def __repr__(self) -> str:
        return f'<Netrc(path="{self.path}", valid={self.is_valid})>'

    @property
    def valid_hosts(self) -> list:
        """ A list of valid netrc machine hosts """
        return [k for k, v in self._valid_hosts.items() if v]

    @property
    def is_valid(self) -> bool:
        """ Checks for a valid netrc file """
        try:
            return self.check_netrc()
        except BrainError:
            return False

    def check_netrc(self) -> None:
        """ Validates the netrc file """

        # check for file existence
        if not self.path.is_file():
            raise BrainError(f'No .netrc file found at {self.path}!')

        # check for correct permissions
        if oct(self.path.stat().st_mode)[-3:] != '600':
            raise BrainError('Your .netrc file does not have 600 permissions. Please fix it by '
                             'running chmod 600 on it. Authentication will not work with '
                             'permissions different from 600.')

        # read the netrc file
        try:
            netfile = netrc.netrc(self.path)
        except netrc.NetrcParseError as nerr:
            raise BrainError(f'Your netrc file was not parsed correctly. Error: {nerr}') from nerr

        # check the netrc file has the allowed SDSS host machines
        nethosts = netfile.hosts.keys()
        badlist = []
        for host in self.allowed_hosts:
            self._valid_hosts[host] = host in nethosts
            if host not in nethosts:
                badlist.append(host)

        # check that the required domains are included
        required = set(['data.sdss.org', 'api.sdss.org']) & set(badlist)
        if required:
            warnings.warn(f"Hosts {', '.join(required)} not found in netrc.  "
                          "You will not be able to remotely access SDSS data.", UserWarning)

        # validate if any are good
        return any(self._valid_hosts.values())

    def check_host(self, host: str) -> bool:
        """ Checks the host against the local netrc file

        Checks if the host is a valid machine in the netrc file

        Parameters
        ----------
        host : str
            The netrc machine name

        Returns
        -------
        bool
            True if the host is valid
        """
        netfile = netrc.netrc(self.path)
        return (host in netfile.hosts.keys() and host in self._valid_hosts and
                self._valid_hosts[host])

    def read_netrc(self, host: str) -> tuple:
        """ Read the netrc file for a given host

        Reads the username, password for the given host machine.  Note this
        returns plaintext username and password.  Do not write out these
        values some place transparent and publicly visible.

        Parameters
        ----------
        host : str
            The netrc machine name

        Returns
        -------
        tuple
            Plain text netrc username, password

        Raises
        ------
        BrainError
            when netrc file fails to pass checks
        ValueError
            when input host is not valid
        """

        if not self.check_netrc():
            raise BrainError('netrc did not pass checks.  Cannot read!')

        if not self.check_host(host):
            raise ValueError(f'{host} must be a valid host in the netrc')

        netfile = netrc.netrc(self.path)
        user, acct, passwd = netfile.authenticators(host)  # pylint: disable=unused-variable
        return user, passwd



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: client.py
# Project: api
# Author: Brian Cherinka
# Created: Friday, 23rd October 2020 3:08:00 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Friday, 23rd October 2020 3:08:00 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import httpx
import re
import warnings
from io import BytesIO
from typing import Union, Type
from sdss_brain import log
from sdss_brain.auth import User
from sdss_brain.api.manager import ApiManager, ApiProfile
from sdss_brain.config import config
from sdss_brain.exceptions import BrainError

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None


apim = ApiManager()


class BaseClient(object):
    """ Base class for creating a client for sending http requests

    This class provides a client for submitting http requests using
    `httpx <https://www.python-httpx.org/>`_.  It provides support for SDSS user
    validation and profile switching between different SDSS APIs.  It provides a
    basic synchronous client, ``SDSSClient``, using
    `httpx Client <https://www.python-httpx.org/advanced/#usage>`_ and asynchronous
    client, ``SDSSAsyncClient``, using `httpx AsyncClient <https://www.python-httpx.org/async/>`_.

    Parameters
    ----------
    route : str, optional
        A full url or route segment, by default None
    user : Union[str, Type[User]], optional
        The name of a valid SDSS user, by default None
    use_api : Union[str, Type[ApiProfile]], optional
        The name of a valid SDSS API profile, by default None
    test : bool, optional
        If True, uses the development API, by default None
    domain : str, optional
        The name of the domain to use for the given API, by default None
    headers : dict, optional
        Any httpx request headers to attach to the client, by default None
    no_progress : bool, optional
        If True, turns off the tqdm progress bar for streaming responses, by default None
    release : str, optional
        The data release to use for the request

    Attributes
    ----------
    user : `~sdss_brain.auth.user.User`
        an SDSS User instance
    api : `~sdss_brain.api.manager.ApiProfile`
        an SDSS API Profile instance
    url: str
        The request url
    client : Union[`httpx.Client`, `httpx.AsyncClient`]
        The underling httpx request client
    response: `httpx.Response`
        The underlying httpx response instance
    data: Union[dict, str, bytes]
        The extracted response content
    """
    _kls: Union[Type[httpx.Client], Type[httpx.AsyncClient]] = None

    def __init__(self, route: str = None, user: Union[str, Type[User]] = None,
                 use_api: Union[str, Type[ApiProfile]] = None, test: bool = None,
                 domain: str = None, headers: dict = None, no_progress: bool = None,
                 release: str = None, **kwargs):
        self.user = None
        self.api = None
        self.url = None
        self.response = None
        self.data = None
        self.no_progress = no_progress
        self.release = release or config.release

        # set the user
        self.set_user(user)

        # set the api when no route or partial route is provided; otherwise use the full url.
        if not route or (route and not route.startswith('http')):
            self.set_api(use_api, domain=domain, test=test)

        # set the URL
        self.set_url(route)

        # set the httpx.Client or httpx.AsyncClient
        self.client = self._kls(headers=headers, **kwargs)  # pylint: disable=not-callable

    def __repr__(self) -> str:
        client = self._kls.__name__
        return (f'<SDSS{client}(api="{str(self.api)}", user="{str(self.user)}")>')

    def set_user(self, user: Union[str, Type[User]]) -> None:
        """ Set the user for the http client

        Sets the SDSS user to use for all remote http requests.

        Parameters
        ----------
        user : Union[str, Type[User]]
            An SDSS username or `~sdss_brain.auth.user.User`

        Raises
        ------
        BrainError
            when the user is not validated
        """
        self.user = User(user) if type(user) == str else user if isinstance(user, User) \
            else config.user
        if not self.user.validated:
            warnings.warn(f"User {user} is not a validated user. Some requests may not work.")

    def set_api(self, use_api: Union[str, Type[ApiProfile]], domain: str = None,
                test: bool = None) -> None:
        """ Set the API profile to use in the http client

        Sets the API profile to use for all remote http requests.

        Parameters
        ----------
        use_api : Union[str, Type[ApiProfile]]
            A API name or `~sdss_brain.api.manager.ApiProfile`
        domain : str, optional
            The domain name to use for the API, by default None
        test : bool, optional
            If True, toggles the development API, by default None

        Raises
        ------
        BrainError
            when no API profile is set
        """
        api = use_api or config.apis.profile
        if isinstance(api, ApiProfile):
            self.api = api
        elif type(api) == str:
            apim.set_profile(api, test=test, domain=domain)
            self.api = apim.profile
        else:
            self.api = None

        if not self.api:
            raise BrainError('No API is set. Set one either by specifying "use_api" on input or '
                             'setting one on the global config.')

    def set_url(self, route: str) -> None:
        """ Sets the url route to use on the http client

        Sets the url to use in the http client for remote requests.  When the input route
        is a full url starting with "http", uses the url as specified.  When the input
        route is a segment, will construct a full url using the base url of the preset SDSS
        API profile.

        Parameters
        ----------
        route : str
            A url or route segment

        Raises
        ------
        BrainError
            when no API profile is set
        """
        if not route:
            return
        elif route.startswith('http'):
            self.url = route
            log.debug(f'Using fully qualified url {route}.')
        else:
            if not self.api:
                raise BrainError(f'No API is set.  Cannot construct a full url for {route}')
            self.url = self.api.construct_route(route)
            log.debug(f'Building url from input route {route} and selected API {self.api}.')

    def _validate_request(self, url: str, method: str) -> None:
        """ Validates some inputs to the request wrapper

        Validates the wrapper request method for a proper url and input
        method type.

        Parameters
        ----------
        url : str
            The request url
        method : str
            The type of http request

        Raises
        ------
        ValueError
            when an invalid method is passed
        ValueError
            when no url is set
        """
        if url:
            self.set_url(url)

        if method not in ['get', 'post', 'stream', 'head']:
            raise ValueError('http request method type can only be "get", "post", '
                             '"head", or "stream".')

        if not self.url:
            raise ValueError('No url set.  Cannot make a request. Please specify a '
                             'url or route segment.')

        # look for any bracketed named arguments in the url
        parts = re.findall(r'{(.*?)}', self.url)
        if parts:
            raise BrainError(f'Request url contains bracket arguments: "{", ".join(parts)}". '
                             'Cannot send request until these are properly replaced.')

    def prepare_data(self, data: dict = None):
        """ Add the SDSS release to input request data

        Automatically adds the client SDSS data release to the input request data
        for cases where the API web server needs it.

        Parameters
        ----------
        data : dict, optional
            any input request data, by default None

        Returns
        -------
        dict
            The request data with release added

        Raises
        ------
        TypeError
            when the input data is not a dict
        """

        # set the release dict and return it when no input data
        release = {'release': self.release}
        if not data:
            return release

        # otherwise check the data type
        if type(data) != dict:
            raise TypeError('input request data is not of type dict')

        # add the release to the request data
        if data and 'release' not in data:
            data.update(release)
        return data

    def _check_response(self, resp: Type[httpx.Response]) -> None:
        """ Checks the returned httpx response

        Checks the httpx response, raises exception for any problem status code,
        and attempts to extract the response content.

        Parameters
        ----------
        resp : Type[httpx.Response]
            The returned httpx response instance

        Raises
        ------
        BrainError
            when the response is not ok
        """
        resp.raise_for_status()
        if resp.is_error:
            raise BrainError('There was an error in the response!')

        self.response = resp
        self._return_data()

    def _return_data(self) -> None:
        """ Extracts the data from a httpx response

        Extracts the relevant httpx response content based on the "content-type"
        from the response header, and sets the data attribute.

        """
        encoding = self.response.headers.get('content-encoding')
        contentype = self.response.headers.get('content-type')

        if 'json' in contentype:
            self.data = self.response.json()
        elif 'text' in contentype:
            self.data = self.response.text
        elif 'octet-stream' in contentype and self.data is None:
            self.data = self.response.content
        else:
            self.data = self.data

    def _create_token_auth_header(self) -> dict:
        """ Create a new request authorization header

        Creates a new http request header containing valid
        Bearer token authorization.

        Returns
        -------
        dict
            A new http request header to be added to the outgoing request

        Raises
        ------
        BrainError
            when no API profile has been set
        ValueError
            when a token cannot be retrieved
        """
        if not self.api:
            raise BrainError("No API profile set. Cannot created token auth header.")

        if self.api.auth_type != 'token':
            return {}

        token = self.api.token
        if not token:
            log.info('No token found.  Attempting to retreive one.')
            token = self.get_token()
            if not token:
                raise ValueError(f'No token retrieved for API {self.api.name}.  Check for a valid '
                                 'user, try self.get_token again, and save your token!')
            else:
                log.info("This token is temporary.  To permanently set one, run self.api.get_token "
                         "and save it in your config file or as an environment variable.")

        return {'Authorization': f'Bearer {token}'}

    def get_token(self) -> str:
        """ Get an API auth token for the specified API

        Retrieves an API auth token for the given API profile, if
        the auth_type for this API is set to "token".  For APIs that only
        use "netrc" authentication, this method is unnecessary.

        Returns
        -------
        str
            A valid API auth token

        Raises
        ------
        ValueError
            when no user is set
        ValueError
            when no API profile is set
        ValueError
            when the user is not properly validated
        """
        if not self.user:
            raise ValueError("No user specified.")

        if not self.api:
            raise ValueError("No API profile specified.")

        if not self.user.validated and not self.user.is_netrc_valid:
            raise ValueError(f'User {self.user} is not properly validated with netrc.  '
                             'Cannot get a token.  Check self.user for proper netrc validation.')

        return self.api.get_token(self.user)


class SDSSClient(BaseClient):
    """ Helper class for sending http requests using httpx.Client"""
    _kls: Type[httpx.Client] = httpx.Client

    def __del__(self):
        if hasattr(self, 'client'):
            self.client.close()

    def __enter__(self):
        ''' Constructor for client use as context manager '''
        return self

    def __exit__(self, type, value, traceback):
        ''' Destructor for client use as context manager '''
        self.client.close()
        return True

    def request(self, url: str = None, data: dict = None, method: str = 'get', files: dict = None,
                content: bytes = None, **kwargs) -> None:
        """ Submit a http request with httpx

        This is a convenience method that wraps `httpx.Client.request`.  It provides support for
        simple "get", "post", or "stream" requests.  For more custom, or complete, control over
        sending requests, use the `client` attribute, and see
        `httpx Client <https://www.python-httpx.org/advanced/#usage>`_ and
        `httpx docs <https://www.python-httpx.org/quickstart/>`_. for more.

        Parameters
        ----------
        url : str, optional
            A url or route segment to send a request to, by default None
        data : dict, optional
            Any data passed along in the request, by default None
        method : str, optional
            The type of http request method, by default 'get'
        files : dict, optional
            Input for multi-part file uploads, by default None
        content : bytes, optional
            Input for binary content, by default None
        kwargs : Any, optional
            Any other httpx request keyword arguments

        Raises
        ------
        BrainError
            when there is an error sending the request
        """

        # validate the input
        self._validate_request(url, method)

        # add any token auth to headers
        headers = None
        if self.api and self.api.auth_type == 'token':
            headers = self._create_token_auth_header()

        # prepare the data with the release
        data = self.prepare_data(data)

        try:
            # try to send the request
            if method == 'stream':
                resp = self._stream_request()
            else:
                params, data, json = (data, None, None) if method == 'get' else (None, data, data)
                resp = self.client.request(method, self.url, params=params, data=data, json=json,
                                           files=files, content=content, headers=headers, **kwargs)

        except httpx.RequestError as exc:
            raise BrainError(f'An error occurred requesting {exc.request.url!r}') from exc
        else:
            # check the response and set the data attribute
            self._check_response(resp)

    def _stream_request(self) -> Type[httpx.Response]:
        """ Stream a httpx response

        Streams response binary content, using ``iter_bytes`` into a
        BytesIO stream object.  For other types of streaming,
        see `<https://www.python-httpx.org/quickstart/#streaming-responses>`_ and use
        the `client` attribute directly.  If ``tqdm`` is installed will display a progress
        bar while streaming content.

        Returns
        -------
        httpx.Response
            The httpx response instance
        """
        # create a BytesIO object
        b = BytesIO()
        with self.client.stream("GET", self.url) as resp:
            # check the status
            resp.raise_for_status()
            # get the content-length
            total = int(resp.headers.get("Content-Length", 0))
            if not tqdm or self.no_progress or total == 0:
                # stream bytes data without progress bar
                for data in resp.iter_bytes():
                    b.write(data)
            else:
                # stream bytes data with progress bar
                with tqdm(total=total, unit_scale=True, unit_divisor=1024, unit="B") as progress:
                    num_bytes_downloaded = resp.num_bytes_downloaded
                    for data in resp.iter_bytes():
                        b.write(data)
                        progress.update(resp.num_bytes_downloaded - num_bytes_downloaded)
                        num_bytes_downloaded = resp.num_bytes_downloaded
            b.seek(0)

        self.data = b
        return resp


class SDSSAsyncClient(BaseClient):
    """ Helper class for sending asynchronous http requests using httpx.AsyncClient"""
    _kls: Type[httpx.AsyncClient] = httpx.AsyncClient

    async def __aenter__(self):
        ''' Constructor for client use as context manager '''
        return self

    async def __aexit__(self, type, value, traceback):
        ''' Destructor for client use as context manager '''
        await self.client.aclose()
        return True

    async def request(self, url: str = None, data: dict = None, method: str = 'get', files: dict = None,
                      content: bytes = None, **kwargs):
        """ Submit a http request with httpx

        This is a convenience method that wraps `httpx.AsyncClient.request`.  It provides support
        for simple "get", "post", or "stream" requests.  For more custom, or complete, control over
        sending requests, use the `client` attribute, and see
        `httpx AsyncClient <https://www.python-httpx.org/async/>`_ and
        `httpx docs <https://www.python-httpx.org/quickstart/>`_. for more.

        Parameters
        ----------
        url : str, optional
            A url or route segment to send a request to, by default None
        data : dict, optional
            Any data passed along in the request, by default None
        method : str, optional
            The type of http request method, by default 'get'
        files : dict, optional
            Input for multi-part file uploads, by default None
        content : bytes, optional
            Input for binary content, by default None
        kwargs : Any, optional
            Any other httpx request keyword arguments

        Raises
        ------
        BrainError
            when there is an error sending the request
        """
        # validate the input
        self._validate_request(url, method)

        # add any token auth to headers
        headers = None
        if self.api and self.api.auth_type == 'token':
            headers = self._create_token_auth_header()

        # prepare the data with the release
        data = self.prepare_data(data)

        try:
            # try to send the request
            if method == 'stream':
                resp = await self._stream_request()
            else:
                params, data, json = (data, None, None) if method == 'get' else (None, data, data)
                resp = await self.client.request(method, self.url, params=params, data=data,
                                                 json=json, files=files, content=content,
                                                 headers=headers, **kwargs)
        except httpx.RequestError as exc:
            raise BrainError(f'An error occurred requesting {exc.request.url!r}') from exc
        else:
            # check the response and set the data attribute
            self._check_response(resp)
        finally:
            await self.client.aclose()

    async def _stream_request(self) -> Type[httpx.Response]:
        """ Stream a httpx response

        Streams response binary content, using ``iter_bytes`` into a
        BytesIO stream object.  For other types of streaming,
        see `<https://www.python-httpx.org/quickstart/#streaming-responses>`_ and use
        the `client` attribute directly.  If ``tqdm`` is installed will display a progress
        bar while streaming content.

        Returns
        -------
        httpx.Response
            The httpx response instance
        """
        # create a BytesIO object
        b = BytesIO()
        async with self.client.stream("GET", self.url) as resp:
            # check the status
            resp.raise_for_status()
            # get the content-length
            total = int(resp.headers.get("Content-Length", 0))
            if not tqdm or self.no_progress or total == 0:
                # stream bytes data without progress bar
                async for data in resp.aiter_bytes():
                    b.write(data)
            else:
                # stream bytes data with progress bar
                with tqdm(total=total, unit_scale=True, unit_divisor=1024, unit="B") as progress:
                    num_bytes_downloaded = resp.num_bytes_downloaded
                    async for data in resp.aiter_bytes():
                        b.write(data)
                        progress.update(resp.num_bytes_downloaded - num_bytes_downloaded)
                        num_bytes_downloaded = resp.num_bytes_downloaded
            b.seek(0)

        self.data = b
        return resp




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: manager.py
# Project: api
# Author: Brian Cherinka
# Created: Friday, 23rd October 2020 3:08:17 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Friday, 23rd October 2020 3:08:18 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import os
import pathlib
import warnings
import yaml
from astropy.table import Table
from functools import wraps
from pydantic import BaseModel, validator, parse_obj_as
from typing import List, Dict
from urllib.parse import urlparse, urlunparse
from sdss_brain import log, cfg_params
from sdss_brain.auth import User
from sdss_brain.api.io import send_post_request
from sdss_brain.exceptions import BrainError

try:
    from tabulate import tabulate
except ImportError:
    tabulate = None


__all__ = ['Domain', 'ApiProfileModel', 'ApiProfile', 'ApiManager', 'apim']


def urljoin(url1: str, url2: str) -> str:
    """ Custom function to join two url paths

    Uses `~urllib.parse.urlparse` and `~urllib.parse.urlunparse`
    to join relevant segments of two urls.  Does not use `~urllib.parse.urljoin`
    as that replaces existing url path with a new path.

    Parameters
    ----------
    url1 : str
        The base url to join to
    url2 : str
        The url segment to join to url1

    Returns
    -------
    str
        A joined url
    """

    e = urlparse(url1)
    t = urlparse(url2)
    final = urlunparse(tuple(strjoin(*z) for z in zip(e, t)))
    return final


def strjoin(str1: str, str2: str) -> str:
    """ Joins two url strings ignoring a leading / """
    if not str2.startswith(str1):
        # use os.path instead of pathlib since it does not trim trailing slashes
        f = os.path.join(str1, str2.lstrip('/')) if str2 else str1
    else:
        f = str2
    return f


# read in the available domains / apis
with open((pathlib.Path(__file__).parent.parent / 'etc/api_profiles.yml').resolve(), 'r') as f:
    profiles = yaml.load(f.read(), Loader=yaml.SafeLoader)
    domains = profiles['domains']
    apis = profiles['apis']


class Domain(BaseModel):
    """ Pydantic class handling validation for SDSS domains """
    name: str
    public: bool = False
    description: str = None

    @validator('name')
    def check_domain_name(cls, value):
        if value != 'localhost' and not value.endswith('sdss.org') and not value.endswith('sdss.utah.edu'):
            raise ValueError(f'Domain name {value} does not fit format of "xxx.sdss.org" or "xxx.sdss.utah.edu"')

        return value

    def __str__(self):
        return str(self.name)

    def __eq__(self, value):
        if type(value) is str:
            return value == self.name
        elif isinstance(value, Domain):
            return value is self


# validate and process the domains yaml section
domains = dict(zip(domains.keys(), parse_obj_as(List[Domain], list(domains.values()))))


def check_domain(func):
    """ Decorator that checks for correct the domain

    Checks that a given input domain is in the list of valid domains for the given
    API profile.  Aso checks for the appropriate inputs when the domain is localhost, i.e.
    for an additional valid port or ngrokid input.  Raises ValueErrors.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        inst, domain = args
        port = kwargs.get('port', None)
        ngrokid = kwargs.get('ngrokid', None)

        if domain not in inst._all_domains:
            raise ValueError(
                f'Input domain "{domain}" not a valid domain/mirror for API profile {inst.name}')

        if (port or ngrokid) and domain != 'local':
            raise ValueError('Domain must be local if a port or ngrokid is set!')

        if domain == 'local' and not (port or ngrokid):
            raise ValueError('A port or ngrokid must be specified when domain is local')

        return func(*args, **kwargs)
    return wrapper


class ApiProfileModel(BaseModel):
    """ Pydantic class handling validation for SDSS API profiles """
    domains: List[str]
    base: str
    mirrors: List[str] = None
    stems: Dict[str, str] = {'test': 'test', 'public': 'public', 'affix': 'prefix'}
    api: bool = False
    routemap: str = None
    auth: Dict[str, str] = {'type': 'netrc', 'route': None}
    description: str = ''
    docs: str = None

    @validator('domains', 'mirrors')
    def domains_in_list(cls, values):
        if not set(values).issubset(set(domains)):
            raise ValueError(f'Not all of the input domains/mirrors are in the list of domains.yml!')
        return values

    @validator('stems')
    def allowed_affixes(cls, values):
        if {'test', 'affix'} - set(values.keys()):
            raise ValueError('stems dictionary must contain at least "test" and "affix" keys!')

        if values['affix'] not in ['prefix', 'suffix']:
            raise ValueError('affix value can only be "prefix" or "suffix"!')
        return values


class ApiProfile(object):
    """ Class representing an API profile

    This class provides an interface for a given SDSS API profile.  It provides
    convenience methods for easily switching domain hosts for a given API, switching
    between production and development paths, and constructing full url route paths.
    It uses the `~urllib.parse.urlparse` URL scheme,
    "scheme://netloc/path;parameters?query#fragment".

    Parameters
    ----------
    name : str
        The name of the API profile
    domain : str, optional
        The name of the domain to use, by default None
    port : int, optional
        The port used for localhost domains, by default None
    ngrokid : int, optional
        The ngrok id used for localhost domains, by default None
    test: bool, optional
        If True, use the development url, by default None

    Attributes
    ----------
        description : str
            A description of the API
        documentation : str
            A url link to any documentation of the API
        auth_type : str
            The type of authentication needed for the API
        domains : dict
            The available domains this API can be accessed on
        mirrors : dict
            The available domains acting as mirrors
        url : str
            The current constructed base API url
        current_domain : str
            The current domain the API it set to use
        name : str
            The name of the API profile
        token: str
            The authenticated token, if any
        info : dict
            A dictionary of information extracted from Pydantic datamodel

    Raises
    ------
    ValueError
        when input name does not match a valid API profile
    """
    def __init__(self, name: str, domain: str = None, port: int = None, ngrokid: int = None,
                 release: str = None, test: bool = None) -> None:

        self.name = name
        # load and validate the name from the list of API profiles
        if name not in apis:
            raise ValueError(f'API Profile {name} not in the apis.yml file. Consider adding it.')
        else:
            self._validated_model = ApiProfileModel.parse_obj(apis[name])
            self.info = self._validated_model.dict()

        self.description = self.info.get('description', '')
        self.documentation = self.info.get('docs', '')

        # load the domains and mirrors
        self.domains = self._get_domains()
        self.mirrors = self._get_domains(mirror=True)
        self._all_domains = {**self.domains} if not self.mirrors else {**self.domains, **self.mirrors}

        # select the current domain
        domain = next(iter(self.domains)) if not domain else domain
        self._select_current_domain(domain, port=port, ngrokid=ngrokid)

        # construct the API url
        self.url = self.construct_url(test=test)

        # set auth type
        self._set_auth_type()

    def __repr__(self) -> str:
        return f'<ApiProfile("{self.name}", current_domain="{self.current_domain}", url="{self.url}")>'

    def __str__(self) -> str:
        return self.name.lower()

    def _set_auth_type(self) -> None:
        """ Sets the API authentication type """
        if self.is_domain_public is True or 'local' in str(self.current_domain):
            self.auth_type = None
        else:
            self.auth_type = self.info['auth'].get('type', 'netrc')

    @property
    def token(self) -> str:
        """ Returns an authentication token """
        return self.check_for_token()

    def check_for_token(self) -> str:
        """ Checks for a proper auth token set as a envvar or in custom config """
        token = f'{self.name.upper()}_API_TOKEN'
        return os.getenv(token) or cfg_params.get(token.lower(), None)

    def construct_token_url(self) -> str:
        """ Construct a login url for requesting tokens """
        # check the API auth_type
        auth = self.info.get('auth', None)
        if auth['type'] != 'token':
            log.info(f'Auth type for API {self.name} is not "token".  No token needed.')
            return

        # get the token route for the given API
        token_route = auth.get('route', None)
        if not token_route:
            raise ValueError(f'No token route specified for API profile {self.name}. '
                             'I do not where to request a token from.')

        return self.construct_route(token_route)

    def get_token(self, user: str) -> str:
        """ Request and receive a valid API auth token

        Requests an auth token for the specified user.  This uses found netrc
        authentication to attempt to request and retrieve a valid token.  The
        token should be saved in an "XXX_API_TOKEN" environment variable or in
        the custom sdss_brain.yml configuration file as "xxx_api_token", where
        "XXX" is the API profile name.

        Parameters
        ----------
        user : str
            The name of the SDSS user

        Returns
        -------
        str
            A valid API auth token

        Raises
        ------
        BrainError
            when the user is not netrc validated
        BrainError
            when a token cannot be extracted from the http response
        """
        if self.token:
            return self.token

        auth = self.info.get('auth', None)
        if auth['type'] != 'token':
            log.info(f'Auth type for API {self.name} is not "token".  No token needed.')
            return

        if type(user) == str:
            user = User(user)
        if not user.validated and not user.is_netrc_valid:
            raise BrainError(f'User {user.name} is not netrc validated!  Cannot access credentials.')

        username, password = user.netrc.read_netrc('api.sdss.org')
        token_url = self.construct_token_url()
        data = send_post_request(token_url, data={'username': username, 'password': password})
        token = data.get('token',
                         data.get('access_token',
                                  data.get('user_token',
                                           data.get('sdss_token', None))))
        if not token:
            raise BrainError('Token request successful but could not extract token '
                             'from response data. Check the returned json response '
                             'for prope key name')
        else:
            tok_name = f'{self.name.upper()}_API_TOKEN'
            log.info(f'Save this token as either a "{tok_name}" environment variable in your '
                     f'.bashrc or as "{tok_name.lower()}" in your custom sdss_brain.yml config file.')
            return token

    @property
    def is_domain_public(self) -> bool:
        """ Checks if current domain is a public one """
        return self.current_domain.public

    @check_domain
    def _select_current_domain(self, domain: str, port: int = None, ngrokid: int = None) -> None:
        if len(self._all_domains) == 1:
            warnings.warn(f'Only one domain available for API profile {self.name}. '
                          'Selecting that one.')
            self.current_domain = list(self._all_domains.values())[0]
            return

        if domain == 'local':
            self.current_domain = self._create_local_domain(port, ngrokid)
        else:
            self.current_domain = self._all_domains.get(domain, None)

    def _get_domains(self, mirror: bool = None) -> dict:
        """ Get the subset of domains valid for the given API

        Set the domains subsets for the specific API.  Sets the
        ``domains`` attribute and the ``mirrors`` attribute.

        Parameters
        ----------
        mirror : bool, optional
            If True, looks for and sets any mirror domains, by default None

        Returns
        -------
        dict
            The set of valid domains for the given API

        Raises
        ------
        ValueError
            when the API profile has no domains entry set
        """
        # get the API profile domains ; add localhost automatically
        key = 'mirrors' if mirror else 'domains'
        profile_domains = self.info.get(key, None)
        if key == 'domains':
            profile_domains.append('local')

        if not mirror and not profile_domains:
            raise ValueError(f'Profile {self.name} has no "domains" entry set!')
        elif mirror and not profile_domains:
            return None
        return {i: domains[i] for i in profile_domains}

    def construct_url(self, test: bool = None, public: bool = None) -> str:
        """ Constructs a new base url

        Constructs a new url given the currently set domain name and the API
        base name.  Can optionally specify a development base using the test or
        public keywords.

        Parameters
        ----------
        test : bool, optional
            If True, add the "test" stem for the development path, by default None
        public : bool, optional
            If True, adds the "public" stem for the deveopment path, by default None

        Returns
        -------
        str
            [description]
        """
        scheme = 'http' if 'local' in self.current_domain else 'https'
        netloc = self.current_domain.name
        path = self._create_base_stem(test, public)
        return urlunparse((scheme, netloc, path, '', '', ''))

    def construct_route(self, route: str) -> str:
        """ Construct a full url to an input route

        Constructs a full url path to the input route.

        Parameters
        ----------
        route : str
            The route component url path

        Returns
        -------
        str
            The full route url

        Example
        -------
        >>> p = ApiProfile('marvin')
        >>> p.url
        'https://sas.sdss.org/marvin/api'
        >>> p.construct_route('general/getroutemap')
        'https://sas.sdss.org/marvin/api/general/getroutemap'
        """
        return urljoin(self.url, route)

    def _create_local_domain(self, port: int, ngrokid: int) -> str:
        """ Create a local domain name """
        if ngrokid:
            domain = f'{ngrokid}.ngrok.io'
        else:
            domain = f'{self.domains["local"].name}:{port}'
        return domain

    @check_domain
    def change_domain(self, domain: str, port: int = None, ngrokid: int = None) -> None:
        """ Change the url domain

        Updates the url "netloc" segment to use the domain name provided.  If the domain is
        "local", also needs either a port number or ngrok id to fully construct a local
        domain name.

        Parameters
        ----------
        domain : str
            The name of the domain to use
        port : int, optional
            The port used for localhost domains, by default None
        ngrokid : int, optional
            The ngrok id used for localhost domains, by default None

        Example
        -------
        >>> p = ApiProfile('marvin')
        >>> p.url
        'https://sas.sdss.org/marvin/api'
        >>> p.change_domains('dr15')
        >>> p.url
        'https://dr15.sdss.org/marvin/api'
        """
        parsed_url = urlparse(self.url)
        params = dict(zip(['scheme', 'netloc', 'path', 'params', 'query', 'fragment'],
                          tuple(parsed_url)))
        if domain == 'local':
            params['scheme'] = 'http'
            params['netloc'] = self._create_local_domain(port, ngrokid)
        else:
            params['scheme'] = 'https'
            params['netloc'] = self._all_domains[domain].name
        self.url = urlunparse(params.values())

        # reset current domain and auth_type
        self.current_domain = self._all_domains[domain]
        self._set_auth_type()

    def _create_base_stem(self, test: bool, public: bool) -> str:
        """ Create a new path stem """
        base = self.info['base']
        api = self.info.get('api', False)
        path = f'{base}'

        if test or public:
            stems = self.info.get('stems', None)
            affix = stems.get('affix', 'prefix')
            testfix = stems.get('test', 'test')
            publicfix = stems.get('public', 'public')
            if test:
                path = f'{testfix}/{path}' if affix == 'prefix' else f'{path}/{testfix}'
            if public:
                path = f'{publicfix}/{path}' if affix == 'prefix' else f'{path}/{publicfix}'

        if api:
            path = f'{path}/api'

        return path

    def change_path(self, test: bool = None, public: bool = None) -> None:
        """ Change the url path

        Updates the url "path" segment.  Called without arguments, with update the path
        to the production base.  If either "test" or "public" is set will update the path
        with the corresponding stems, used to switch to development urls.

        Parameters
        ----------
        test : bool, optional
            If True, add the "test" stem for the development path, by default None
        public : bool, optional
            If True, adds the "public" stem for the deveopment path, by default None

        Example
        -------
        >>> p = ApiProfile('marvin')
        >>> p.url
        'https://sas.sdss.org/marvin/api'
        >>> p.change_path(test=True)
        >>> p.url
        'https://sas.sdss.org/test/marvin/api'
        """
        path = self._create_base_stem(test, public)
        parsed_url = urlparse(self.url)
        params = dict(zip(['scheme', 'netloc', 'path', 'params', 'query', 'fragment'],
                          tuple(parsed_url)))
        params['path'] = path
        self.url = urlunparse(params.values())


class ApiManager(object):
    """ Class for managing SDSS APIs

    This class provides an interface for handling and managing the selection of various
    SDSS APIs.  It allows toggling of the current API to use for all remote
    requests.

    Attributes
    ----------
    domains : dict
        A dictionary of available SDSS domains
    apis : dict
        A dictionary of available SDSS APIs
    profile : Type[ApiProfile]
        The currently selected API to use
    """
    def __init__(self) -> None:
        self.domains = domains
        self.apis = {a: ApiProfile(a) for a in apis.keys()}
        self.profile = None

    def __repr__(self) -> str:
        return (f'<ApiManager(current_api="{str(self.profile)}", n_domains="{len(self.domains)}", '
                f'n_apis="{len(self.apis)}")>')

    def list_apis(self) -> list:
        """ List the available SDSS APIs

        Displays the complete list of available SDSS APIs.

        Returns
        -------
        list
            The list of available SDSS APIs
        """
        return list(self.apis.values())

    def list_domains(self) -> list:
        """ List the available SDSS domains

        Displays the complete list of available SDSS domain names.

        Returns
        -------
        list
            The list of available SDSS domains
        """
        return list(self.domains.values())

    def set_profile(self, name: str, domain: str = None, test: bool = None) -> None:
        """ Set the current API profile

        Sets the current API to the named profile

        Parameters
        ----------
        name : str
            The name of the API
        domain: str, optional
            The name of the domain to switch to, by default None
        test: bool, optional
            If True, sets the API profile to development, by default None
        """

        if name not in self.apis:
            raise ValueError(f'Input profile {name} not an available SDSS API.')

        self.profile = self.apis.get(name, None)

        if domain:
            self.profile.change_domain(domain)

        if test:
            self.profile.change_path(test=test)

    def identify_api_from_url(self, url: str) -> tuple:
        """ Identify and extract an API and domain type from a URL.

        Identifies the type of API and domain name the input URL
        is using.  Loops over all available API and domain profiles in the
        ApiManager checks against extracted url parts from `~urllib.parse.urlparse`.

        Parameters
        ----------
        url : str
            The full url string

        Returns
        -------
        tuple
            The identified API profile and domain name

        Raises
        ------
        ValueError
            when the url does not start with http
        """
        if not url.startswith('http'):
            raise ValueError('Input url does not start with http.')

        api = domain = None
        # extract url parts
        parts = urlparse(url)
        # identify the API
        for v in self.apis.values():
            if v.info['base'] in parts.path:
                api = v.name

        # identify the domain
        for k, d in self.domains.items():
            if d.name == parts.netloc:
                domain = k

        return api, domain

    def display(self, value: str, pprint: bool = False, show_docs: bool = True,
                show_desc: bool = True, **kwargs) -> Table:
        """ Display the APIs or domains as an Astropy Table

        Display the list of available SDSS APIs or domains as an
        Astropy Table.

        Parameters
        ----------
        value : str
            Either "api(s)" or "domain(s)"
        pprint : bool, optional
            If True, pretty print the Astropy Table, by default False
        show_docs : bool, optional
            If True, include the "docs" column in the API table, by default True
        show_desc : bool, optional
            If True, include the "description" column in the API table, by default True
        kwargs: Any
            Other kwargs for Table.pprint (pretty print) method

        Returns
        -------
        `~astropy.table.Table`
            An Astropy Table of information

        Raises
        ------
        TypeError
            when the input value is not a string
        ValueError
            when the input value is not either "apis" or "domains"
        """
        if type(value) != str:
            raise TypeError('Value can only be a string')

        value = value.lower()
        if value not in ['apis', 'api', 'domains', 'domain']:
            raise ValueError('Value can only be "apis" or "domains"!')

        rows = []
        cols = []
        if 'domain' in value:
            # create display columns
            cols = ['key', 'name', 'public', 'description']

            # create table rows
            for k, v in self.domains.items():
                row = v.dict()
                row['key'] = k
                rows.append(row)
        elif 'api' in value:
            # create display columns
            cols = ['key', 'base', 'description', 'domains', 'mirrors', 'auth', 'docs']
            if not show_docs:
                cols.remove('docs')
            if not show_desc:
                cols.remove('description')

            # create table rows
            for k, v in self.apis.items():
                row = {kk: vv for kk, vv in v.info.items() if kk in cols}
                row['domains'] = ', '.join(row['domains'])
                row['mirrors'] = ', '.join(row['mirrors']) if row['mirrors'] else ''
                row['auth'] = row['auth']['type']
                row['key'] = k
                rows.append(row)

        table = Table(rows, names=cols)

        if pprint:
            table.pprint(**kwargs)
            return

        return table

    def generate_rst_table(self, value: str, show_docs: bool = True,
                           show_desc: bool = True, **kwargs) -> str:
        """ Generate a rst-formatted table

        Generates an rst-formatted table of the available domains or APIs
        using the `tabulate <https://github.com/astanin/python-tabulate>`_
        python package.  This method is good for dropping tables into Sphinx
        documentation.

        Parameters
        ----------
        value : str
            Either "api(s)" or "domain(s)"
        show_docs : bool, optional
            If True, include the "docs" column in the API table, by default True
        show_desc : bool, optional
            If True, include the "description" column in the API table, by default True
        kwargs: Any
            Other kwargs for the tabulate method

        Returns
        -------
        str
            The rst formatted table as a string

        Raises
        ------
        ImportError
            when the tabulate package is not installed.
        """
        if not tabulate:
            raise ImportError('package tabulate not found.  Cannot generate rst table.')

        # create the table as a numpy array of data
        table = self.display(value, show_docs=show_docs, show_desc=show_desc).as_array()

        # convert the table into an rst table using tabulate
        rst = tabulate(table, headers='keys', tablefmt='rst', disable_numparse=True, **kwargs)

        return rst


apim = ApiManager()



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: config.py
# Project: sdss_brain
# Author: Brian Cherinka
# Created: Friday, 14th February 2020 1:41:34 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Tuesday, 17th March 2020 4:20:37 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
from sdss_brain import cfg_params, log, tree
from sdss_brain.auth import User
from sdss_brain.api.manager import ApiManager
from sdss_brain.exceptions import BrainError


class Config(object):
    """ Main configuration class for SDSS Brain

    This class helps set up global configurations for the ``Brain`` to use.

    Attributes
    ----------
    download : bool
        When True, automatically downloads all data
    ignore_db : bool
        When True, automatically ignores all database connections
    work_versions: dict
        A dictionary containing the survey version numbers to work with non-released data
    user: type[User]
        A validated SDSS user
    """

    def __init__(self):
        self._mode = 'auto'
        self._release = None
        self.download = False
        self.ignore_db = False
        self.work_versions = None

        # load default config parameters
        self._load_defaults()

        # get allowed releases from the Tree
        self._allowed_releases = tree.get_available_releases()

        # set a default release or get the latest one
        default_release = self._custom_config.get('default_release', None)
        self.release = default_release or self._get_latest_release()

        # set any work versions from the config
        self.set_work_versions()

        # set default sdss user
        self.set_user()

        # set default API
        self.set_api()

    def __repr__(self):
        return f'<SDSSConfig(release={self.release}, mode={self.mode})>'

    @property
    def mode(self) -> str:
        """ The mode of operation for the Brain """
        return self._mode

    @mode.setter
    def mode(self, value: str):
        assert value in ['local', 'remote', 'auto'], ('config.mode must be "local",'
                                                      '"remote", or "auto".')
        self._mode = value

    @property
    def release(self) -> str:
        """ The SDSS public, internal, or "work" release """
        return self._release

    @release.setter
    def release(self, value: str) -> None:
        value = value.upper()
        if value not in self._allowed_releases:
            raise BrainError(f'trying to set an invalid release version {value}. '
                             f'Valid releases are: {", ".join(self._allowed_releases)}')

        # if work release or IPL, check for a validated user
        if 'dr' not in value.lower() and not self.user.validated:
            raise BrainError(f'User {self.user} is not validated. Can only access public data. '
                             'Cannot access "work" data or internal SDSS releases.  Consider '
                             'validating the user, or check your SDSS netrc credentials.')

        # replant the tree
        if value.lower() == 'work':
            tree.replant_tree('sdsswork')
        else:
            tree.replant_tree(value.lower())

        self._release = value

    def set_release(self, version: str = None) -> None:
        """ Set a new release

        Set a new global data release to use.  Can be either
        a public, internal SDSS data release, or can use non-released data
        by setting release to "WORK".

        Parameters
        ----------
        version : str, optional
            The new data release to set, by default uses the latest public DR
        """
        if not version:
            version = self._get_latest_release()
            log.info(f'Setting release to latest: {version}')
        self.release = version

    def _get_latest_release(self) -> str:
        """ Get the latest public DR release """
        drsonly = [i for i in self._allowed_releases if 'DR' in i]
        return max(drsonly, key=lambda t: int(t.rsplit('DR', 1)[-1]))

    def list_allowed_releases(self) -> list:
        """ List the allowed releases based on the available tree environment configurations """
        return self._allowed_releases

    def _load_defaults(self) -> None:
        """ Load the Brain config yaml file and update any parameters """

        # update any matching Config values
        for key, value in cfg_params.items():
            if hasattr(self, key):
                self.__setattr__(key, value)

        self._custom_config = cfg_params

    def set_work_versions(self, values: dict = {}) -> None:
        """ Set the versions used for sdsswork

        Sets the versions used by sdswork globally into the config.
        Input is a valid dictionary containing version names and numbers as
        key value pairs, e.g. `{'drpver':'v2_4_3', 'run2d':'v1_1_1', 'apred':'r12'}`.
        Optionally can set them permanently in the custom configuration YAML file.

        The input dictionary is merged with any values specified from the custom config.

        Parameters
        ----------
        values : dict, optional
            Dictionary of version names:numbers needed in paths, by default {}

        Raises
        ------
        ValueError
            when the input is not a dictionary
        """

        if type(values) != dict:
            raise ValueError('input versions must be a dict')

        cfg_work = self._custom_config.get('work_versions', {})
        cfg_work.update(values)
        self.work_versions = cfg_work

    def set_user(self, user: str = 'sdss', password: str = None) -> None:
        """ Set a new global user

        Sets a new global `~sdss_brain.auth.user.User`.  By default with be
        set to the generic "sdss" user.  Can override the default by setting
        the "default_username" and "default_userpass" keywords in the custom
        YAML configuration file.

        Parameters
        ----------
        user : str, optional
            The username to use, by default 'sdss'
        password : str, optional
            The password to use to validate the user, by default None.

        Raises
        ------
        BrainError
            when a user cannot be validated
        """

        default_user = self._custom_config.get('default_username', None)
        default_pass = self._custom_config.get('default_userpass', None)
        user = default_user or user
        password = default_pass or password
        log.debug(f'Setting user {user}. '
                  '{"No password specified." if not password else "Password specified."}')
        self.user = User(user)
        if not self.user.validated and user and password:
            self.user.validate_user(password)

            if not self.user.validated:
                raise BrainError(f'Could not validate default user {user}!')
            else:
                log.debug(f'Validated user {user}')

        if not self.user.validated:
            log.warning(f'User {user} is not validated.  Check your netrc credentials '
                        'or validate your user with config.set_user(username, password)')

    def set_api(self, name: str = None, domain: str = None, test: bool = None) -> None:
        """ Set a new global API to use

        Sets the `~sdss_brain.api.manager.ApiManager` and a new global
        `~sdss_brain.api.manager.ApiProfile`.  Can set permanently by setting
        the "default_api" keywords in the custom YAML configuration file.

        Parameters
        ----------
        name : str, optional
            The name of a valid SDSS API, by default None
        domain : str, optional
            The name of the domain to set on the API profile, by default None
        test : bool, optional
            If True, uses the development API, by default None
        """
        default_api = self._custom_config.get('default_api', None)
        name = name or default_api
        self.apis = ApiManager()
        if name:
            self.apis.set_profile(name, domain=domain, test=test)


config = Config()




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: __init__.py
# Project: auth
# Author: Brian Cherinka
# Created: Wednesday, 21st October 2020 2:51:55 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Wednesday, 21st October 2020 2:51:55 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

from .netrc import *
from .htpass import *
from .user import *



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: cubes.py
# Project: tools
# Author: Brian Cherinka
# Created: Sunday, 11th October 2020 3:15:17 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Sunday, 11th October 2020 3:15:17 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import numpy as np
import warnings
from astropy.io import fits
from sdss_brain.tools import Spectrum
from sdss_brain.helpers import sdss_loader


class Cube(Spectrum):
    """ Base class for cubes """
    pass


@sdss_loader(name='mangacube', defaults={'wave': 'LOG'}, mapped_version='manga:drpver',
             pattern=r'(?P<plateifu>(?P<plate>\d{4,5})-(?P<ifu>\d{3,5}))|(?P<mangaid>\d{1,2}-\d{4,9})')
class MangaCube(Spectrum):
    """ Class representing a MaNGA IFU datacube for a single galaxy """
    specutils_format: str = 'MaNGA cube'
    _api = ('marvin', 'cubes/{plateifu}/')

    def _load_object_from_db(self):

        # do nothing if not connected to the database
        if not self.db.connected:
            return

        # issue warning if no model or models attached to database handler
        if not self.db.model or not self.db.models:
            warnings.warn('No model(s) found in db handler.  Cannot query for a cube.'
                          'Try passing in an explicit ORM or setting the schema and model with '
                          'db.load_model or db.load_schema.')
            return

        # make a database call to retrieve the Cube row
        self.data = self.db.session.query(self.db.model).join(self.db.models.IFUDesign,
                                                              self.db.models.PipelineInfo,
                                                              self.db.models.PipelineVersion).\
            filter(self.db.model.plateifu == self.plateifu,
                   self.db.models.PipelineVersion.version == self.drpver).one_or_none()

        if not self.data:
            warnings.warn(f"No data returned from database for {self.plateifu}")
            return

        # extract header info and reformat to proper fits.Header
        # TODO move this into the datamodel product
        head = [(k, int(v) if v.isnumeric() else float(v) if v.replace('.', '').isnumeric(
        ) else v.strip(), self.data.header.comments[i]) for i, (k, v) in enumerate(self.data.header.items())]
        self.header = fits.Header(head)

        # TODO - slow; 5 seconds to access all 3 from local db; improve this?
        flux = self.data.get3DCube('flux')
        ivar = self.data.get3DCube('ivar')
        mask = self.data.get3DCube('mask')

        # construct a new astropy FITS HDUList
        hdulist = fits.HDUList([fits.PrimaryHDU(data=None),
                                fits.ImageHDU(flux, name='FLUX', header=self.header),
                                fits.ImageHDU(ivar, name='IVAR'), fits.ImageHDU(mask, name='MASK'),
                                fits.ImageHDU(self.data.wavelength.wavelength, name='WAVE')])

        # load the spectrum
        self._load_spectrum(hdulist)

    def _load_object_from_api(self):

        # do nothing if no remote client
        if not self.remote:
            return

        # submit the remote request
        self.remote.client.request(method='post', data={'release': self.release})

        # extract data from the response
        self.data = self.remote.client.data['data']
        self.mangaid = self.data['mangaid']
        self.ra = self.data['ra']
        self.dec = self.data['dec']
        self.redshift = self.data['redshift']
        self.wavelength = np.array(self.data['wavelength'])

        # set the header
        # TODO - fix header - see above
        header = fits.Header.fromstring(self.data['header'])
        head = [(k, int(v) if v.isnumeric() else float(v) if v.replace('.', '').isnumeric(
        ) else v.strip(), header.comments[i]) for i, (k, v) in enumerate(header.items())]
        self.header = fits.Header(head)

        # example to retrieve flux for central spaxel
        # TODO - look into improving speed of 3d flux retrieval
        # TODO - improve url extension-ing
        self.shape = np.array(self.data['shape'])
        x_cen, y_cen = self.shape // 2
        spaxel_url = self.remote.extend_url(f'quantities/{x_cen}/{y_cen}/')
        self.remote.client.request(spaxel_url, data={'release': self.release})

        # construct a FITS hdulist
        flux = np.zeros(shape=((len(self.wavelength),) + tuple(self.shape)))
        ivar = np.zeros(shape=((len(self.wavelength),) + tuple(self.shape)))
        mask = np.zeros(shape=((len(self.wavelength),) + tuple(self.shape)))
        flux[:, x_cen, y_cen] = np.array(self.remote.client.data['data']['flux']['value'])
        ivar[:, x_cen, y_cen] = np.array(self.remote.client.data['data']['flux']['ivar'])
        mask[:, x_cen, y_cen] = np.array(self.remote.client.data['data']['flux']['mask'])
        hdulist = fits.HDUList([fits.PrimaryHDU(data=None),
                                fits.ImageHDU(flux, name='FLUX', header=self.header),
                                fits.ImageHDU(ivar, name='IVAR'), fits.ImageHDU(mask, name='MASK'),
                                fits.ImageHDU(self.wavelength, name='WAVE')])

        # load the spectrum
        self._load_spectrum(hdulist)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: parsing.py
# Project: helpers
# Author: Brian Cherinka
# Created: Wednesday, 7th October 2020 10:54:07 am
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Wednesday, 7th October 2020 10:54:07 am
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import re
import pathlib
import warnings
from typing import Match, Pattern, Union
from itertools import groupby
from sdss_brain import log


def create_object_pattern(regex: Union[str, Pattern] = None, keys: list = None, keymap: dict = None,
                          delimiter: str = '-', exclude: list = None, include: list = None,
                          order: list = None) -> str:
    """ Create a regex pattern to parse data input by

    Parameters
    ----------
        regex : str
            A custom regex pattern
        keys : list
            A list of (access) names to build a pattern out of
        keymap : dict
            A dict of key name and pattern to build a pattern out of
        delimiter : str
            The delimiter to use when joining the keys.  Default is "-".
        exclude : list
            A list of names to exclude from the keys
        include : list
            A list of names to only include from the keys
        order : list
            A list of names specifying the order in which to create the keyed pattern

    Returns
    -------
        pattern : str
            A regex pattern to use for parsing an objectid
    """

    # use a custom regex pattern
    if regex:
        pattern = rf'(?P<objectid>(?![/$.])({regex}))'
        return pattern

    # if no keys or keymap, use a greedy default
    if not keys and not keymap:
        pattern = r'(?P<objectid>^[^/$.](.+)?)'
        return pattern

    assert keys or keymap, 'Either a list of keys or a keymap must be specified.'
    if not isinstance(keys, (list, type(None))) or not isinstance(keymap, (dict, type(None))):
        raise TypeError('keys must be list; keymap must be a dict')
    if not keys and keymap:
        keys = list(keymap.keys())

    # make a copy of the original key order
    keys_copy = keys.copy()

    # exclude the named keys
    if exclude:
        keys = list(set(keys) - set(exclude))

    # only include the named keys
    if include or order:
        good = order or include
        keys = list(set(good) & set(keys))

    # resort the keys by the original key order
    keys.sort(key=lambda i: keys_copy.index(i))

    # order the keys
    if order:
        keys.sort(key=lambda i: order.index(i))

    patts = []
    for k in keys:
        if not keymap:
            patts.append(fr'(?P<{k}>(.+)?)')
        else:
            patts.append(fr'(?P<{k}>{keymap[k]})')

    # join into a single pattern
    delimiter = '-' if not delimiter else delimiter
    pattern = rf'(?P<objectid>(?![/$.])({delimiter.join(patts)}))'

    return pattern


def parse_data_input(value: str, regex: str = None, keys: list = None, keymap: dict = None,
                     delimiter: str = '-', exclude: list = None, include: list = None,
                     order: list = None, inputs: bool = False) -> dict:
    ''' Parse data input for a filename or an object id

    Parameters
    ----------
        value : str
            The input string to perform a pattern match on
        regex : str
            A custom regex pattern
        keys : list
            A list of (access) names to build a pattern out of
        keymap : dict
            A dict of key name and pattern to build a pattern out of
        delimiter : str
            The delimiter to use when joining the keys.  Default is "-".
        exclude : list
            A list of names to exclude from the keys
        include : list
            A list of names to only include from the keys
        order : list
            A list of names specifying the order in which to create the keyed pattern
        inputs : bool
            If True, returns the parser inputs.  Default is False.

    Returns
    -------
        matches : dict
            A dict with keys "filename", "objectid", and any other matches

    Example
    -------
        >>> # parse a filename
        >>> parse_data_input('/path/to/a/file.txt')
            {'filename': '/path/to/a/file.txt', 'objectid': None, 'parsed_groups': None}

        >>> # parse an objectid as is
        >>> parse_data_input('8485-1901')
            {'filename': None, 'objectid': '8485-1901', 'parsed_groups': ['8485-1901', '485-1901']}

        >>> # parse an objectid using a custom pattern
        >>> parse_data_input('8485-1901', regex=r'(?P<plate>\d{4,5})-(?P<ifu>\d{3,5})')
            {'filename': None, 'objectid': '8485-1901', 'plate': '8485', 'ifu': '1901', 'parsed_groups': ['8485-1901', '8485', '1901']}

        >>> # parse an objectid using access keywords
        >>> keys=['drpver', 'plate', 'ifu', 'wave']
        >>> parse_data_input('v1-8485-1901-LOG', keys=keys)
            {'filename': None, 'objectid': 'v1-8485-1901-LOG', 'drpver': 'v1', 'plate': '8485',
             'ifu': '1901', 'wave': 'LOG', 'parsed_groups': ['v1-8485-1901-LOG', 'v1', '8485', '1901', 'LOG']}

        >>> # parse an objectid specifying the input order of the keys
        >>> parse_data_input('8485-1901', keys=keys, order=['plate', 'ifu'])
            {'filename': None, 'objectid': '8485-1901', 'plate': '8485', 'ifu': '1901', 'parsed_groups': ['8485-1901', '8485', '1901']}

    '''

    if not isinstance(value, (str, pathlib.Path)):
        return ValueError('input value must be a str or pathlib.Path')

    # check if regex has named groups
    # is_named = re.findall(r'\?P<(.*?)>', regex) if regex else None

    # check if regex is a compiled pattern already; if so, strip out string pattern
    if regex and type(regex) == re.Pattern:
        regex = regex.pattern

    # set default file pattern
    file_pattern = r'(?P<filename>^[/$.](.+)?(.[a-z]+))'

    # create an object id regex pattern using a specified pattern or generate a default one
    obj_pattern = create_object_pattern(regex=regex, keys=keys, keymap=keymap, delimiter=delimiter,
                                        exclude=exclude, include=include, order=order)

    # final pattern
    pattern = fr'^{file_pattern}|{obj_pattern}$'

    # store pattern inputs
    input_patts = {'pattern': pattern, 'input_regex': regex,
                   'object_pattern': obj_pattern, 'file_pattern': file_pattern}

    # compile and match the patterm
    comp_pattern = re.compile(pattern)
    pattern_match = re.match(comp_pattern, str(value))

    # if no match, assume value is a filename and return nothing
    if not pattern_match:
        warnings.warn('No pattern match found.  Defaulting to input value as a filename.')
        if inputs:
            return {'filename': value, 'parsed_inputs': input_patts}
        return {'filename': value}

    # check for named group, then any groups, then a match without groups
    matches = pattern_match.groupdict() or pattern_match.groups() or pattern_match.group()

    # add the groups to a new key (remove None and duplicate values)
    matches['parsed_groups'] = [k for k, _ in groupby(sorted(pattern_match.groups(),
                                                             key=lambda x: pattern_match.groups().index(x))) if k] \
        if not matches.get('filename', None) else None

    # store the parser inputs
    if inputs:
        matches['parsed_inputs'] = input_patts

    return matches


def raw_parse(value: str, regex: str = None) -> Match:
    ''' Match a string via a regex pattern with no frills

    Parameters
    ----------
        value : str
            The input string to match on
        regex : str
            The regex pattern to use for matching

    Returns
    -------
        A matched group

    '''
    pattern = re.compile(regex)
    match = re.match(pattern, value)
    if not match:
        return None

    return match.groupdict() or match.groups() or match.group()
